""" WhichBarcodeCheckMeta Class """

import tkinter
from kalmus.tkinter_windows.meta_info_windows.MetaInfoWindow import MetaInfoWindow
from kalmus.tkinter_windows.gui_utils import resource_path


class WhichBarcodeCheckMeta():
    """
    WhichBarcodeCheckMeta Class
    GUI window for user to choose which barcode to check with the meta information
    """
    def __init__(self, barcode_1, barcode_2, barcodes_stack):
        """
        Initialize

        :param barcode_1: The barcode 1
        :param barcode_2: The barcode 2
        :param barcodes_stack: The dictionary that stored all the barcode in memory
        """
        self.barcode_1 = barcode_1
        self.barcode_2 = barcode_2
        self.barcode_stacks = barcodes_stack

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Which Barcode to Check Meta Info")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Option variable
        self.barcode_option = tkinter.StringVar(self.window)
        self.barcode_option.set("Barcode 1")

        # Option radio button
        radio_barcode_1 = tkinter.Radiobutton(self.window, text="Barcode 1", variable=self.barcode_option,
                                              value="Barcode 1")
        radio_barcode_1.grid(row=0, column=0, padx=50)
        radio_barcode_1.select()

        radio_barcode_2 = tkinter.Radiobutton(self.window, text="Barcode 2", variable=self.barcode_option,
                                              value="Barcode 2")
        radio_barcode_2.grid(row=1, column=0, padx=50)

        # Check button
        self.button_check = tkinter.Button(master=self.window, text="Check", command=self.check_meta_info)
        self.button_check.grid(row=2, column=0, padx=50)

    def check_meta_info(self):
        """
        Get the selected option from user and instantiate the MetaInfoWindow using the given option (barcode)
        """
        # Get the selection
        which_barcode = self.barcode_option.get()

        # Quit current window
        self.window.destroy()

        # Instantiate the MetaInfoWindow using the selection
        if which_barcode == "Barcode 1":
            MetaInfoWindow(self.barcode_1, self.barcode_stacks)
        elif which_barcode == "Barcode 2":
            MetaInfoWindow(self.barcode_2, self.barcode_stacks)



""" Barcode Generator Class"""

import numpy as np
import json

from kalmus.barcodes.Barcode import ColorBarcode, BrightnessBarcode

# Available metrics for computing the color of a frame
color_metrics = ["Average", "Median", "Mode", "Top-dominant", "Weighted-dominant",
                 "Brightest", "Bright"]

# Available types of sampling frame (which part of frame is region of interest)
frame_types = ["Whole_frame", "High_contrast_region", "Low_contrast_region", "Foreground", "Background"]

# Available types of barcode
barcode_types = ["Color", "Brightness"]


def build_barcode_from_json(path_to_json, barcode_type="Color"):
    """
    Helper function that build a barcode object from the attributes stored in a json file

    :param path_to_json: Path to the json file
    :param barcode_type: Type of the barcode that stored in the json file
    :return: The barcode built from the json file at given path
    """
    assert barcode_type in barcode_types, "Invalid barcode type. The available types of " \
                                          "the barcode are {:s}".format(str(barcode_types))
    with open(path_to_json, "r") as infile:
        object_dict = json.load(infile)
    infile.close()

    # Make sure the barcode_type is correctly capitalized
    barcode_type = barcode_type.capitalize()

    if barcode_type == "Color":
        barcode = ColorBarcode(color_metric=object_dict["color_metric"], frame_type=object_dict["frame_type"],
                               sampled_frame_rate=object_dict["sampled_frame_rate"],
                               skip_over=object_dict["skip_over"], total_frames=int(object_dict["total_frames"]),
                               barcode_type=barcode_type)

        barcode.colors = np.array(object_dict["colors"]).astype("uint8")

    elif barcode_type == "Brightness":
        barcode = BrightnessBarcode(color_metric=object_dict["color_metric"], frame_type=object_dict["frame_type"],
                                    sampled_frame_rate=object_dict["sampled_frame_rate"],
                                    skip_over=object_dict["skip_over"], total_frames=int(object_dict["total_frames"]),
                                    barcode_type=barcode_type)

        barcode.brightness = np.array(object_dict["brightness"]).astype("uint8")

    barcode.set_letterbox_bound(object_dict["low_bound_ver"], object_dict["high_bound_ver"],
                                object_dict["low_bound_hor"], object_dict["high_bound_hor"])

    if "meta_data" in object_dict.keys():
        barcode.meta_data = object_dict["meta_data"]

    barcode.barcode = np.array(object_dict["barcode"])

    barcode.video = None

    if "fps" in object_dict.keys():
        if object_dict["fps"] is not None:
            barcode.fps = float(object_dict["fps"])

    barcode.film_length_in_frames = int(object_dict["film_length_in_frames"])

    if "save_frames_in_generation" in object_dict.keys():
        if object_dict["save_frames_in_generation"]:
            barcode.save_frames_in_generationa = object_dict["save_frames_in_generation"]
            barcode.saved_frames = np.array(object_dict["saved_frames"])

    return barcode


class BarcodeGenerator():
    """
    Barcode Generator Class
    """
    def __init__(self, frame_type="Whole_frame", color_metric="Average", barcode_type="Color",
                 sampled_frame_rate=1, skip_over=0, total_frames=10):
        """
        Initialize the parameters for the barcode generator

        :param frame_type: The type of the frame sampling
        :param color_metric: The metric of computing the frame color
        :param barcode_type: The type of the generated barcode
        :param sampled_frame_rate: The frame sample rate \
        (one frame will be sampled from every sampled_frame_rate frames)
        :param skip_over: How many frames to skip with at the beginning of the input video
        :param total_frames: Total number of frames that will be computed (included in the barcode/sampled frames)
        """
        assert frame_type in frame_types, "Invalid frame acquisition method. Five types of frame acquisition" \
                                          " methods are available including {:s}".format(str(frame_types))
        assert color_metric in color_metrics, "Invalid color metric. Seven color metrics are available " \
                                              "including {:s}".format(str(color_metrics))
        assert barcode_type in barcode_types, "Invalid barcode type. Two types of barcode are available " \
                                              "including {:s}".format(str(barcode_types))
        assert not (color_metric == "Bright" and frame_type in frame_types[1:]), \
            "Color metric Bright can not be used when the frame acquisition " \
            "methods are {:s}".format(str(frame_types[1:]))

        self.frame_type = frame_type
        self.color_metric = color_metric
        self.barcode_type = barcode_type
        self.sampled_frame_rate = sampled_frame_rate
        self.skip_over = skip_over
        self.total_frames = total_frames
        self.barcode = None

    def instantiate_barcode_object(self):
        """
        Instantiate the barcode object using the given generation parameters
        """
        if self.barcode_type == "Color":
            self.barcode = ColorBarcode(self.color_metric, self.frame_type, self.sampled_frame_rate,
                                        self.skip_over, self.total_frames, barcode_type="Color")
        elif self.barcode_type == "Brightness":
            self.barcode = BrightnessBarcode(self.color_metric, self.frame_type, self.sampled_frame_rate,
                                             self.skip_over, self.total_frames, barcode_type="Brightness")

    def generate_barcode(self, video_file_path, user_defined_letterbox=False,
                         low_ver=-1, high_ver=-1, left_hor=-1, right_hor=-1,
                         num_thread=None, save_frames=False, rescale_frames_factor=-1,
                         save_frames_rate=4):
        """
        Generate the barcode

        :param video_file_path: The path to the video file
        :param user_defined_letterbox: Whether use the user defined the letterbox, or use the \
        automatically found letterbox
        :param low_ver: The lower vertical letterbox given by user
        :param high_ver: The higher vertical letterbox given by user
        :param left_hor: The left horizontal letterbox given by user
        :param right_hor: The right horizontal letterbox given by user
        :param num_thread: Number of thread for computation. None == Single thread. num_thread > 1: multi-thread
        :param save_frames: Whether to save the frames during the barcode generation
        :param rescale_frames_factor: factor to rescale the input frames during the generation
        """
        self.instantiate_barcode_object()
        if user_defined_letterbox:
            self.barcode.set_letterbox_bound(up_vertical_bound=high_ver, down_vertical_bound=low_ver,
                                             left_horizontal_bound=left_hor, right_horizontal_bound=right_hor)
        if save_frames and save_frames_rate > 0:
            self.barcode.enable_save_frames(sampled_rate=save_frames_rate)

        if rescale_frames_factor > 0:
            self.barcode.enable_rescale_frames_in_generation(rescale_frames_factor)

        if self.barcode_type == "Color":
            if num_thread is not None:
                self.barcode.multi_thread_collect_colors(video_file_path, num_thread)
            else:
                self.barcode.collect_colors(video_file_path)
        elif self.barcode_type == "Brightness":
            if num_thread is not None:
                self.barcode.multi_thread_collect_brightness(video_file_path, num_thread)
            else:
                self.barcode.collect_brightness(video_file_path)

    def generate_barcode_from_json(self, json_file_path, barcode_type=None):
        """
        Generate the barcode from a json file, which contain a dictionary representation of barcode object

        :param json_file_path: the path to the json file
        :param barcode_type: the type of the barcode saved in the json file
        """
        if barcode_type is None:
            barcode_type = self.barcode_type
        self.barcode = build_barcode_from_json(json_file_path, barcode_type=barcode_type)

    def get_barcode(self):
        """
        return the barcode object stored in the Barcode generator

        :return: The generated barcode
        """
        assert self.barcode is not None, "There is not a generated barcode"
        return self.barcode



""" LoadStackWindow Class """

import tkinter
from kalmus.tkinter_windows.gui_utils import update_graph, resource_path
import copy


class LoadStackWindow():
    """
    LoadStackWindow Class
    GUI window for user to load the barcode from the memory into the main window of the kalmus
    """
    def __init__(self, barcode_stack, barcode_1, barcode_2, axes, canvas):
        """
        Initialize

        :param barcode_stack: The dictionary that stores all the barcode on memory
        :param barcode_1: The barcode 1
        :param barcode_2: The barcode 2
        :param axes: The plotted figure axes in the main window
        :param canvas: The plotted figure canvas in the main window
        """
        self.barcode_stack = barcode_stack
        self.barcode_1 = barcode_1
        self.axes = axes
        self.canvas = canvas
        self.barcode_2 = barcode_2

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))
        self.window.wm_title("Barcodes on Memory Stack")

        # Set up the list box that shows all the barcode on memory using their keys in the dictionary
        self.listbox = tkinter.Listbox(self.window, selectmode=tkinter.SINGLE, width=65, height=20)
        self.listbox.grid(row=0, column=0)

        # List all the barcodes with their keys
        for barcode_names in self.barcode_stack.keys():
            self.listbox.insert(tkinter.END, barcode_names)

        # Button to load the barcode from the memory
        self.button_load = tkinter.Button(master=self.window, text="Load Selected Barcode", command=self.load_stack)
        self.button_load.grid(row=3, column=0, sticky=tkinter.W)

        # Option variable that stores the position which the new barcode from json will replace with
        self.barcode_option = tkinter.StringVar(self.window)
        self.barcode_option.set("Barcode 1")

        # Radio buttons for the replaced barcode selection
        radio_barcode_1 = tkinter.Radiobutton(self.window, text="Barcode 1", variable=self.barcode_option,
                                              value="Barcode 1", anchor='w')
        radio_barcode_1.grid(row=1, column=0, sticky=tkinter.W)
        radio_barcode_1.select()

        radio_barcode_2 = tkinter.Radiobutton(self.window, text="Barcode 2", variable=self.barcode_option,
                                              value="Barcode 2", anchor='w')
        radio_barcode_2.grid(row=2, column=0, sticky=tkinter.W)

    def load_stack(self):
        """
        Load the barcode from the memory and use it to replace one of the displayed barcodes in the main window
        """
        # Get the selection of the barcode on memory
        barcode_key = str(self.listbox.get(self.listbox.curselection()))

        # Get which barcode in the main window to replace
        which_barcode = self.barcode_option.get()

        # Replace with the barcode in the main window
        if which_barcode == "Barcode 1":
            self.barcode_1.__dict__ = copy.deepcopy(self.barcode_stack[barcode_key].__dict__.copy())
            self.barcode_1.__class__ = self.barcode_stack[barcode_key].__class__
        elif which_barcode == "Barcode 2":
            self.barcode_2.__dict__ = copy.deepcopy(self.barcode_stack[barcode_key].__dict__.copy())
            self.barcode_2.__class__ = self.barcode_stack[barcode_key].__class__

        # Clear the axes
        self.axes[0][0].cla()
        self.axes[1][0].cla()
        self.axes[0][1].cla()
        self.axes[1][1].cla()

        # Always plot the barcode with longer width below
        if self.barcode_1.get_barcode().shape[1] > self.barcode_2.get_barcode().shape[1]:
            temp = copy.deepcopy(self.barcode_1)

            self.barcode_1.__dict__ = self.barcode_2.__dict__.copy()
            self.barcode_2.__dict__ = temp.__dict__.copy()

        # Update the graph/plotted figure in the main window
        update_graph(barcode_1=self.barcode_1, barcode_2=self.barcode_2, axes=self.axes)

        # Redraw the canvas
        self.canvas.draw()

        # Quit the window
        self.window.destroy()



from kalmus.barcodes.BarcodeGenerator import BarcodeGenerator
import argparse
import sys


def parse_args_into_dict(args):
    ap = argparse.ArgumentParser(description="Command line Barcode generator")
    # Video path
    ap.add_argument("-p", "--path", required=True, type=str, help="Path to the video")

    # Color metric
    ap.add_argument("--color_metric", required=True, type=str,
                    help="Color metric. available options: [Average, Median, Mode, Top-dominant, Weighted-dominant, "
                         "Brightest, Bright]")

    # Frame type
    ap.add_argument("--frame_type", required=True, type=str,
                    help="Type of the frame sampling. available options: [Whole_frame, High_contrast_region, "
                         "High_contrast_region, Low_contrast region, Foreground, Background]")

    # Barcode type
    ap.add_argument("--barcode_type", required=True, type=str,
                    help="Type of the generated barcode. available options: [Color, Brightness]")

    # Skip over how many frames after collecting colors/brightness
    ap.add_argument("--skip", required=False, type=int, help="Skip number of frames before collecting colors",
                    default=0)

    # Sampled frame rate
    ap.add_argument("-s", "--step", required=True, type=int, help="Sampled frame rate",
                    default=1)

    # Total frames in barcode
    ap.add_argument("-t", "--total_frames", required=False, type=int,
                    help="Total number of frames to be collected from the input video into the barcode",
                    default=10)

    # Total frames in barcode
    ap.add_argument("--num_thread", required=False, type=int,
                    help="Number of threads used to generate barcode. Single by default.",
                    default=None)
    # Total frames in barcode
    ap.add_argument("--saved_frame_rate", required=False, type=float,
                    help="Saved one frame every saved_frame_rate seconds. No frames saved during generation "
                         "by default",
                    default=None)
    # Rescale frame in generation
    ap.add_argument("--rescale_frame_factor", required=False, type=float,
                    help="Factor that frame will be resized during the generation. Frame is not resized by default",
                    default=-1)
    # Total frames in barcode
    ap.add_argument("-o", "--output_path", required=False, type=str,
                    help="Path to the output JSON barcode. By default, path==saved_{barcode_type}_{frame_type}_"
                         "{color_metric}.json",
                    default=None)
    return vars(ap.parse_args(args))


def main(args=sys.argv[1:]):
    print(sys.argv)
    args = parse_args_into_dict(args=args)

    if args["saved_frame_rate"] is None:
        saved_frame_in_generation = False
    else:
        saved_frame_in_generation = True
    barcode_generator = BarcodeGenerator(frame_type=args["frame_type"], color_metric=args["color_metric"],
                                         barcode_type=args["barcode_type"], sampled_frame_rate=args["step"],
                                         skip_over=args["skip"], total_frames=args["total_frames"])
    barcode_generator.generate_barcode(video_file_path=args["path"], num_thread=args["num_thread"],
                                       save_frames=saved_frame_in_generation, save_frames_rate=args["saved_frame_rate"],
                                       rescale_frames_factor=args["rescale_frame_factor"])
    barcode = barcode_generator.get_barcode()
    barcode.save_as_json(filename=args["output_path"])



""" Utility Artist """

import numpy as np
import cv2
import csv
from scipy import ndimage
from skimage.morphology import disk, remove_small_objects
from skimage.segmentation import watershed
from skimage.filters import rank, sobel
from skimage.future import graph
import scipy.stats as stats


def compute_dominant_color(image, n_clusters=3, max_iter=10, threshold_error=1.0, attempts=10):
    """
    Compute the dominant color of an input image using the kmeans clustering. The centers of the
    clusters are the dominant colors of the input image

    :param image: input image. Either a multi-channel color image or a grayscale image (2D image) \
    Expected shape of image is height x width x (channels)
    :param n_clusters: number of clusters
    :param max_iter: maximum iterations before terminating kmeans clustering
    :param threshold_error: threshold error for terminating kmeans clustering. kmean clustering terminate \
    when the errors between the current computed and previous computed cluster center is under this \
    threshold
    :param attempts: Number of attempts to rerun the Kmeans clustering with different initial cluster \
    centers. Since kmeans clustering randomly choose n number of cluster in its initialization process.
    :return: An array of n cluster centers and an array of relative size (in percentage) of the clusters. \
    Expected output shape: n_clusters x channels (cluster centers), n_clusters (relative size in percentage) \
    E.g. for an input color image with 3 channels and n_clusters = 3  Output can be [[255, 255, 255], \
    [126, 75, 198], [186, 145, 122]], [0.4, 0.5, 0.1]
    """
    assert len(image.shape) >= 2, "The input must be a 2 dimensional image"
    # Flatten the image
    # If the image is in grayscale
    image = flatten_image(image)
    # make sure the dtype of the flatten image is np.float32
    image = np.float32(image)
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, max_iter, threshold_error)
    ret, label, centers = cv2.kmeans(image, n_clusters, None, criteria, attempts, cv2.KMEANS_RANDOM_CENTERS)

    percent_of_dominance = compute_percents_of_labels(label)

    return centers, percent_of_dominance


def flatten_image(image):
    """
    Flat the input 2D image into an 1D image while preserve the channels of the input image
    with shape==[height x width, channels]

    :param image: Input 2D image (either multi-channel color image or greyscale image)
    :return: The flatten 1D image. shape==(height x width, channels)
    """
    assert len(image.shape) >= 2, "The input image must be a 2 Dimensional image"
    if len(image.shape) == 3:
        image = image.reshape((-1, image.shape[2]))
    elif len(image.shape) == 2:
        image = image.reshape((-1, 1))
    return image


def compute_percents_of_labels(label):
    """
    Compute the ratio/percentage size of the labels in an labeled image

    :param label: the labeled 2D image
    :return: An array of relative size of the labels in the image. Indices of the sizes in the array \
    is corresponding to the labels in the labeled image. E.g. output [0.2, 0.5, 0.3] means label 0's size \
    is 0.2 of the labeled image, label 1' size is 0.5 of the labeled image, and label 2's size is 0.3 of \
    the labeled image.
    """
    # Get the bins of the histogram. Since the last bin of the histogram is [label, label+1]
    # We add 1 to the number of different labels in the labeled image when generating bins
    num_labels = np.arange(0, len(np.unique(label)) + 1)
    # Histogramize the label image and get the frequency array percent_of_dominance
    (percent_of_dominance, _) = np.histogram(label, bins=num_labels)
    # Convert the dtype of frequency array to float
    percent_of_dominance = percent_of_dominance.astype("float")
    # Normalized by the sum of frequencies (number of pixels in the labeled image)
    percent_of_dominance /= percent_of_dominance.sum()
    return percent_of_dominance


def compute_mode_color(image, bin_size=10):
    """
    compute the mode color of an input image

    :param image: either a multi-channel color image or a single channel greyscale image \
    Expected shape of image: height x width x (channels)
    :param bin_size: Histogramize the input image. Color/intensity of each pixel in the \
    image will be accumulate in the bins with size==bin_size. The output mode color/intensity \
    is always an integer multiple of the bin_size.
    :return: The mode color of the image (modes of the channels), shape=channels, and counts of the \
    mode colors happened in the input image, shape==channels
    """
    image = flatten_image(image)
    image = image // bin_size
    mode_color, counts = stats.mode(image, axis=0)
    return (mode_color[0] * bin_size), counts


def compute_mean_color(image):
    """
    Compute the average/mean color of the input multi-channel image or greyscale image.

    :param image: input image. Either a multi-channel color image or a single channel greyscale image
    :return: The average color of the image (averaged across the channels). shape==channels.
    """
    image = flatten_image(image)
    avg_color = np.average(image, axis=0)
    return avg_color


def compute_median_color(image):
    """
    Compute the median color of the input multi-channel color image or a single channel greyscale image.

    :param image: input image. Either a multi-channel color image or a single channel greyscale image
    :return: The median color of the image (median values of channels), shape==channels
    """
    image = flatten_image(image)
    median_color = np.median(image, axis=0)
    return median_color


def compute_brightest_color_and_brightness(grey_image, color_image, return_min=False,
                                           gaussian_blur=False, blur_radius=15):
    """
    Find the brightest pixel in an image and return the color and brightness at that pixel

    :param blur_radius: The radius of the gaussian filter
    :param gaussian_blur: Whether to apply a gaussian filter before finding the brightest point the grey image
    :param grey_image: The greyscale image. single channel 2D image. Expected shape==(row/height, col/width)
    :param color_image: The corresponding color image. Expected shape==(row/height, col/width, channels)
    :param return_min: If true return the color and brightness of the darkest pixel as well
    :return: The color and brightness of the brightest pixel (, color and brightness of the darkest pixel)
    """
    if gaussian_blur:
        grey_image = cv2.GaussianBlur(grey_image.copy(), (blur_radius, blur_radius), 0)
    min_brightness, max_brightness, min_loc, max_loc = cv2.minMaxLoc(grey_image)
    if return_min:
        return color_image[max_loc[::-1]], max_brightness, max_loc[::-1], \
               color_image[min_loc[::-1]], min_brightness, min_loc[::-1]
    else:
        return color_image[max_loc[::-1]], max_brightness, max_loc[::-1]


def find_bright_spots(image, n_clusters=3, blur_radius=21, amount_of_bright_parts=0.8, return_all_pos=False):
    """
    Find the indices location of the top-k brightest spots in an color image.

    :param image: input image. Must be an mutli-channel RGB color image
    :param n_clusters: expected number of clusters/brightest spots in the input image
    :param blur_radius: radius of the Gaussian blur kernel that used to smooth the image
    :param amount_of_bright_parts: amount of bright parts in an image. Used to find the lower bound for \
    distinguishing the bright and non-bright part of the input image. Range of amount_of_bright_parts is \
    in [0, 1] (all non-bright -> all bright)
    :return:
    """
    assert amount_of_bright_parts >= 0 and amount_of_bright_parts <= 1, "Range of the sample ration is in [0, 1]"
    assert n_clusters >= 1, "The number of bright spots must be larger or equal to 1"

    amount_of_bright_parts = amount_of_bright_parts * 100
    # Convert BRG to Greyscale
    grayscale_img = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)

    # Blur the image with radius = blurRadius
    blurred_img = cv2.GaussianBlur(grayscale_img, (blur_radius, blur_radius), 0)

    # Compute the lower bound threshold
    lower_bound = np.percentile(blurred_img, 100 - amount_of_bright_parts) - 1

    # Threshold the imgae by setting any pixel with value larger than lower bound to 255
    threshed_img = cv2.threshold(blurred_img, lower_bound, 255, cv2.THRESH_BINARY)[1]

    # Purifiy the edges of the brightest spots.
    threshed_img = cv2.erode(threshed_img, None, iterations=2)
    threshed_img = cv2.dilate(threshed_img, None, iterations=4)

    # Get the location of all white pixel in binary threshold image
    locs = np.argwhere(threshed_img == 255)
    try:
        # convert to np.float32
        locs_above_threshold = np.float32(locs)

        # define criteria and apply kmeans()
        criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 10, 1.0)
        ret, label, center = cv2.kmeans(locs_above_threshold, n_clusters, None, criteria, 10, cv2.KMEANS_RANDOM_CENTERS)

        # Compute the percentage of each clusters in the image.
        percent_of_dominance = compute_percents_of_labels(label)

        if return_all_pos:
            return label, locs_above_threshold.astype("int32"), percent_of_dominance
        else:
            return center.astype("int32"), percent_of_dominance

    except:
        # Catch exception return negative array.
        if return_all_pos:
            return np.array([-1]), np.array([-1]), np.array([-1])
        else:
            return np.ones(shape=(n_clusters, 1)) * -1, np.ones(shape=(n_clusters, 1)) * -1


def random_sample_pixels(img, sample_ratio=0, mode="row-col"):
    """
    Randomly sample an amount of pixels from an image based on the given sampled ratio.
    Two sampling mode are available.
    row-col: sampling first across the row and then across the column on each sampled row
    The output sampled image is still 2D and keep same aspect ratio as the input image, which
    can be used to visualize the sampling effect on the image.
    flat: sampling across the flat input image. The shape and aspect ratio of the input image
    are not preserved due to the flatten process, but it sample the pixels much faster than
    'row-col' mode. The distribution of the sampling result is similar to that from the 'row-col'

    :param img: Input 2D image. Either a multi-channel color image or a single channel greyscale image \
    Expected shape==height x width x (channels)
    :param sample_ratio: The amount of pixels sampled from the image. in range [0, 1]
    :param mode: two sampling mode are available. \
    1) 'row-col' sampling mode 2) 'flat' sampling mode
    :return: If mode="flat", return the resampled array of pixels (1-d flat array of data points) \
             If mode="row-col", return the resampled image
    """
    assert 0 <= sample_ratio <= 1, "The sample ratio is in the range of [0, 1]"
    if sample_ratio == 1:
        return img
    elif sample_ratio == 0:
        return np.array([]).astype(img.dtype)
    if mode == "row-col":
        # To keep the aspect ratio of the input image, sample equal ratio on the columns and rows
        ratio = (sample_ratio) ** (1 / 2)
        # Number of row to sample
        row_size = int(img.shape[0] * ratio)
        # Number of column to sample
        col_size = int(img.shape[1] * ratio)
        # Sample the row first
        random_row_indices = np.sort(np.random.choice(np.arange(img.shape[0]), row_size, replace=False))
        random_row = img[random_row_indices]
        random_pixels = []
        # Then, in each row, sampled a number of pixels/columns
        for row in random_row:
            random_col_indices = np.sort(np.random.choice(np.arange(img.shape[1]), col_size, replace=False))
            random_pixels.append(row[random_col_indices])
        random_pixels = np.array(random_pixels)
    elif mode == "flat":
        # Flatten the image
        flat_img = flatten_image(img)
        # Generate the possible indices
        indices = np.arange(flat_img.shape[0])
        # Generate the sampled indices
        sampled_indices = np.random.choice(indices, int(sample_ratio * flat_img.shape[0]), replace=False)
        # Sampled the flat image using the sampled indices
        random_pixels = flat_img[sampled_indices]
    else:
        # If none of two modes is specified
        print("Invalid sampling mode. Two sampling options are 'row-col' and 'flat'")
        return
    return np.array(random_pixels)


def watershed_segmentation(image, minimum_segment_size=0.0004, base_ratio=0.5, denoise_disk_size=5,
                           gradiant_disk_size=5,
                           marker_disk_size=15):
    """
    Label the connected regions in an image.
    Use edge detection approach with watershed algorithm.
    adjust the critical gradient (edge gradient) with the distribution of the input image's
    intensity gradient .

    :param grey_image: The input image for region segmentation
    :param minimum_segment_size: The minimum size of the segments in the segmented image in percentage \
    (size is the relative ratio of the image) The segmented regions smaller than the minimum size will be merged \
    to its neighboring larger segmented components
    :param base_ratio: Amount of regions at least in the image. The image in watershed transformation is \
    differentiated into two parts regions + boundaries. The base ratio is the ratio between the least amount of \
    pixels that are regions and the total amount of pixels in the image.
    :param denoise_disk_size: the kernel size of the denoise median filter
    :param gradiant_disk_size: the kernel size of the gradient filter that is used for determining the amount of \
    boundaries
    :param marker_disk_size: the kernel size of the gradient filter that is used for generating transformation \
    marker
    :return: the segmented image, where shape==input_image.shape. and regions are labeled from 0 to n-1, where \
    n is the number of regions in the labeled image. Functions also return the greyscale image corresponding to \
    the original image
    """
    assert 0 <= minimum_segment_size < 1, "The minimum size of the segments (in percentage ratio) is in range [0, 1)"
    # Gray Scale image
    grey_image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)

    # denoise image
    denoised = rank.median(grey_image, disk(denoise_disk_size))

    num_of_pixels = image.shape[0] * image.shape[1]

    # local gradient
    gradient = rank.gradient(denoised, disk(gradiant_disk_size))
    critical = np.sort(gradient, axis=None)

    # Use std to adjust the ratio of the regions in the image
    criteria = np.std(gradient)
    # A high std of the image's intensity gradient means the intensities of image are changed more rapidly
    # and frequently, so there are more boundaries but less regions. A low std of the image's intensity
    # gradients means the changes across the image are more smooth. Such coherence indicate a larger ratio
    # of regions in the image
    # Divide adjust ratio by the std of the gradient distribution
    adjust_ratio = 0.25 * 25 / (criteria + 1e6)
    if (adjust_ratio > 0.45):
        adjust_ratio = 0.45
    # Assume at least 50% of an image is the content (at most 50% of an image
    # is edge), and at most 95% of an image is the content (at least 5% of an
    # image is edge)
    # Critical is the critical threshold of intensity gradient that separate the regions and boundaries
    # Pixels with gradients larger than critical are considered as boundaries
    # Pixels with gradients smaller than critical are considered as regions
    critical = critical[int(num_of_pixels * (base_ratio + adjust_ratio))]

    # Minimum size of a segment
    # Segments under the minimum size will be joined into larger segments
    segment_min_size = int(num_of_pixels * minimum_segment_size)
    # find continuous region which marked by gradient lower than critical gradient
    markers = rank.gradient(denoised, disk(marker_disk_size)) < critical
    markers = ndimage.label(remove_small_objects(markers, segment_min_size))[0]

    # process the watershed transformation
    regions = watershed(gradient, markers)

    # Return the labeled image and the corresponding greyscale image
    return regions, grey_image


def get_rag(gray_image, labels):
    """
    Get the region adjacency graph using the labeled image and corresponding the edge map generated by
    greyscale image with sobel filter

    :param gray_image: The greyscale image corresponding to the labeled image
    :param labels: a labeled segmented image, where the pixels in the image are labeled with index of the \
    segmented regions.
    :return: The region adjacency graph in dictionary \
    see https://scikit-image.org/docs/stable/auto_examples/segmentation/plot_rag_boundary.html for more \
    references
    """
    # Get rag of the labeled image
    edge_map = sobel(gray_image)
    rag = graph.rag_boundary(labels, edge_map)
    return rag


def rag_to_matrix(rag, num_regions):
    """
    Transfer the region adjacency dictionary to a more accessible region adjacency matrix
    where in the matrix, 1 means the region of corresponding column index is adjacent to the
    region of corresponding row index/
    e.g.
    0,1
    1,0 means region 0 and region 1 are adjacent

    :param rag: region adjacency dictionary
    :param num_regions: number of regions in the region adjacency graph
    :return: An binary adjacency matrix with shape==num_regions x num_regions
    """
    matrix = np.zeros((num_regions, num_regions), dtype=np.int8)
    for i in range(1, num_regions + 1):
        elements = rag[i]
        adict = dict(elements)
        adj_list = adict.keys()
        for keys in adj_list:
            matrix[i - 1][keys - 1] = 1
    return matrix


def color_of_regions(labels, original_image):
    """
    Compute average color and brightest color of the regions and record
    the relative size of the regions as a ratio with respect to the size
    of whole image.

    :param labels: The labeled image. Expected shape==height x width. Integer labels
    :param original_image: The original color image corresponding to the label image
    :return: A list of average color of the regions, a list of brightest color of the regions, and \
    a list of sizes of the regions. The order of the regions in list is the same as they are in \
    labeled image
    """
    # Average Color of the region
    avg_colors_list = []

    # Brightest Color of the region
    brightest_colors_list = []

    # Sizes of the regions in the image.
    region_sizes = []

    # Size of the whole image in pixels
    num_of_pixels = original_image.shape[0] * original_image.shape[1]

    # Set of labels in the image
    labels_set = np.unique(labels)
    # Greyscale image
    grey_image = cv2.cvtColor(original_image, cv2.COLOR_RGB2GRAY)

    for region in labels_set:
        indices = labels == region

        # Extract the color info of regions from the original image
        extract_image = original_image[indices]
        region_sizes.append(len(extract_image) / num_of_pixels)
        average_color = np.average(extract_image, axis=0)
        avg_colors_list.append(average_color)

        # Extract the brightness related info of regions from the greyscale image
        extract_gray_image = grey_image[indices]
        brightest_color_loc = np.argmax(extract_gray_image)
        brightest_color = extract_image[brightest_color_loc]
        brightest_colors_list.append(brightest_color)

    return avg_colors_list, brightest_colors_list, region_sizes


def contrast_between_regions(region_colors, matrix, region_weights=None):
    """
    Compute the contrast between the segmented regions in image using the color of regions
    and adjacency matrix

    :param region_colors: A list of colors of segmented regions
    :param matrix: A 2D adjacency matrix that describe the spatial relationship between the regions \
    Expect a binary matrix, where 1 means adjacent and 0 means non-adjacent
    :param region_weights: A 1D array of weights that can be applied onto the contrast calculate for each \
    regions. \
    e.g. Regions may have different sizes, and you can weight the computed contrast by the size of the \
    regions. Expected shape of the regions weight is shape==number of regions
    :return: A 2D numpy matrix where each entry is the contrast between the row indexed region and column \
    indexed region. Contrast ratio >= 1, an entry with 0 means row indexed region and column indexed region \
    are non-adjacent
    """
    assert matrix.shape[0] == matrix.shape[1], "Invalid shape of the adjacency matrix, the matrix must" \
                                               "be a 2D binary numpy array with a square shape. shape[0]" \
                                               "==shape[1]"
    assert matrix.shape[0] == len(region_colors), "Incompatible adjacency matrix with region_colors," \
                                                  " different number of regions in matrix and region_colors" \
                                                  " list"
    # List of row that contains the contrast between two regions
    contrast_list = []
    # If no region weights are given, then use the uniform weights
    if region_weights is None:
        region_weights = np.ones(shape=len(matrix))

    # Compute the contrast between each adjacent regions
    for i in range(len(matrix)):
        contrast_row = []
        for j in range(len(matrix[0])):
            # If two regions are adjacent
            if matrix[i][j] != 0:
                # Compute the contrast and append it to the list
                contrast = contrast_ratio(region_colors[i], region_colors[j]) * region_weights[i]
                contrast_row.append(contrast)
            else:
                # Else append 0 as non-adjacent
                contrast_row.append(0)
        contrast_list.append(contrast_row)
    contrast_list = np.array(contrast_list)

    return contrast_list


def _RGB2sRGB(RGB):
    """
    Convert the 24-bits Adobe RGB color to the standard RGB color defined
    in Web Content Accessibility Guidelines (WCAG) 2.0
    see https://www.w3.org/TR/2008/REC-WCAG20-20081211/#relativeluminancedef for more references

    :param RGB: The input RGB color or colors shape== ... x 3 (R, G, and B channels)
    :return: converted sRGB colors
    """
    sRGB = RGB / 255
    return sRGB


def _sRGB2luminance(sRGB):
    """
    Compute luminance of standard RGB color
    see https://www.w3.org/TR/2008/REC-WCAG20-20081211/#relativeluminancedef for more references

    :param sRGB: Input sRGB color or colors shape==... x 3 (sR, sG, and sB channels)
    :return: the luminance computed using the sRGB color
    """
    lRGB = np.zeros(shape=sRGB.shape)
    lRGB[sRGB <= 0.03928] = sRGB[sRGB <= 0.03928] / 12.92
    lRGB[sRGB > 0.03928] = ((sRGB[sRGB > 0.03928] + 0.055) / 1.055) ** 2.4
    luminance = 0.2126 * lRGB[..., 0] + 0.7152 * lRGB[..., 1] + 0.0722 * lRGB[..., 2]
    return luminance


def contrast_ratio(color1, color2):
    """
    Compute the contrast ratio between two 24-bits RGB colors in range [0, 255]
    see https://www.w3.org/TR/2008/REC-WCAG20-20081211/#contrast-ratiodef for more references

    :param color1: one of the 24-bits RGB colors in range [0, 255]
    :param color2: one of the 24-bits RGB colors in range [0, 255]
    :return: the contrast ratio between two 24-bits RGB colors. Contrast ratio >= 1
    """
    luminance1 = _sRGB2luminance(_RGB2sRGB(color1))
    luminance2 = _sRGB2luminance(_RGB2sRGB(color2))
    ratio = (luminance1 + 0.05) / (luminance2 + 0.05)
    if ratio < 1:
        ratio = 1 / ratio
    return ratio


def get_contrast_matrix_and_labeled_image(frame, minimum_segment_size=0.0004):
    """
    Helper function that use the watershed method to segment the input image return the matrix of the
    brightness contrast of each segmented region with respect to its neighbors (adjacent segmented regions)

    :param frame: The input frame
    :param minimum_segment_size: The minimum size of the segmented region in the ratio to the whole frame. Range (0, 1)
    :return: The matrix with shape (num_regions x num_regions) whose cell [i, j] represents the contrast \
             between the region i and region j, and the corresponding labeled image (segmentation)

    """
    labels, grey_frame = watershed_segmentation(frame, minimum_segment_size=minimum_segment_size)
    try:
        adjacency_matrix = rag_to_matrix(get_rag(grey_frame, labels), len(np.unique(labels)))
        avg_color, _, region_sizes = color_of_regions(labels, frame)
        contrast_matrix = contrast_between_regions(avg_color, adjacency_matrix)
    except:
        contrast_matrix = np.array([[0]])
        labels = np.ones(shape=frame.shape)
    return contrast_matrix, labels


def grabcut_foreback_segmentation(image, start_row=0, start_col=0, row_size=-1, col_size=-1, num_iter=3,
                                  return_masks=False):
    """
    Perform the GrabCut segmentation over the input image with a rectangle of possible foreground
    specified by user. The GrabCut segment the image into two parts foreground and background, and
    the function return the 1D image of the foreground and background as the output

    :param image: The input image for GrabCut segmentation
    :param start_row: The starting row of the foreground rectangle of possible foreground
    :param start_col: The starting col of the foreground rectangle of possible foreground
    :param row_size: The vertical length of the rectangle
    :param col_size: The horizontal length of the rectangle
    :param num_iter: The number of iterations for GrabCut to run
    :param return_masks: Return the foreground and background boolean mask if True \
                         Return the 1D image of foreground pixels and 1D image of background pixels if False \
                         False by default
    :return: If return_masks is False (default), 1D image of the foreground part of the image, and 1D image of \
             the background part of the image Expected shape== Number of pixels x channels. \
             If return_masks is True, return boolean masks foreground and background. Expected shape==image.shape
    """
    if start_row < 0:
        start_row = 0
    if start_col <= 0:
        start_col = image.shape[1] // 6
    if row_size <= 0:
        row_size = image.shape[0] - 1
    if col_size <= 0:
        col_size = image.shape[1] * 2 // 3

    mask = np.zeros(image.shape[:2], np.uint8)
    # Temporary array for background Gaussian mixture model
    background_model = np.zeros((1, 65), np.float64)
    # Temporary array for foreground Gaussian mixture model
    foreground_model = np.zeros((1, 65), np.float64)
    # Rectangle box for possible foreground
    rectangle = (start_col, start_row, col_size, row_size)
    cv2.grabCut(image, mask, rectangle,
                background_model, foreground_model,
                num_iter, cv2.GC_INIT_WITH_RECT)
    if return_masks:
        # The obvious foreground and probable foreground are marked with 1 and 3
        # The obvious background and probable background are marked with 0 and 2
        return (mask == 1) | (mask == 3), (mask == 0) | (mask == 2)
    else:
        # The obvious foreground and probable foreground are marked with 1 and 3
        foreground_image = image[(mask == 1) | (mask == 3)]
        # The obvious background and probable background are marked with 0 and 2
        background_image = image[(mask == 0) | (mask == 2)]

    return foreground_image, background_image


def find_letter_box_from_videos(video, num_sample=30):
    """
    Find the position of letterbox (black bars around the scene) in the input cv2 video object.
    The function samples out num_sample (20 by default) number of frames from cv2 video object, and find the position
    of letterbox in each frame. The function uses the median results from all sampled frames as the final position of
    video's letterbox.

    Notice that the function assumes the letterbox is black or very close to black.

    :param video: Input video object captured by cv2.VideoCapture()
    :param num_sample: Number of frames to sample from video for finding letterbox
    :return: The smaller row index of letterbox (bound for upper horizontal letterbox), \
             the larger row index of letterbox (bound for lower horizontal letterbox), \
             the smaller col index of letterbox (bound for left vertical letterbox), \
             and the larger col index of letterbox (bound for right vertical letterbox). \
             \
             The results are the median results on num_sample number of frames \
             \
             To extract the frame without letterbox from letterboxing frame, \
    """
    film_length_in_frames = int(video.get(cv2.CAP_PROP_FRAME_COUNT))

    # Sampled out num_sample number of frames from input video
    possible_indexes = np.arange(film_length_in_frames // 6, film_length_in_frames * 5 // 6, 1)
    frame_indexes = np.random.choice(possible_indexes, num_sample, replace=True)

    # list of possible smaller row indexes (bounding the upper letterbox)
    possible_low_bound_ver = []
    # list of possible larger row indexes (bounding the lower letterbox)
    possible_high_bound_ver = []
    # list of possible smaller column indexes (bounding the left letterbox)
    possible_low_bound_hor = []
    # list of possible larger column indexes (bounding the right letterbox)
    possible_high_bound_hor = []

    # Find the letterbox position on each sampled frame
    for frame_index in frame_indexes:
        video.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
        success, frame = video.read()
        if success:
            low_bound_v, high_bound_v, low_bound_h, high_bound_h = get_letter_box_from_frames(frame)
            possible_low_bound_ver.append(low_bound_v)
            possible_high_bound_ver.append(high_bound_v)
            possible_low_bound_hor.append(low_bound_h)
            possible_high_bound_hor.append(high_bound_h)

    # Take the medians as final results
    low_bound_ver = int(np.median(possible_low_bound_ver))
    high_bound_ver = int(np.median(possible_high_bound_ver))
    low_bound_hor = int(np.median(possible_low_bound_hor))
    high_bound_hor = int(np.median(possible_high_bound_hor))

    return low_bound_ver, high_bound_ver, low_bound_hor, high_bound_hor


def get_letter_box_from_frames(frame, threshold=5):
    """
    Find the position of letterbox (black bars around the scene) in the input cv2 video object.
    The function assumes the letter box of the frame is black (dark)

    :param frame: Input frame
    :param threshold: The brightness threshold value that distinguish \
                      the region of interest (bright) and letterbox (dark)
    :return: The smaller row index of letterbox (bound for upper horizontal letterbox), \
             the larger row index of letterbox (bound for lower horizontal letterbox), \
             the smaller col index of letterbox (bound for left vertical letterbox), \
             and the larger col index of letterbox (bound for right vertical letterbox). \
             \
             To extract the frame without letterbox from letterboxing frame, \
    """
    low_bound_ver = 0
    high_bound_ver = frame.shape[0]
    low_bound_hor = 0
    high_bound_hor = frame.shape[1]

    for i in range(0, frame.shape[0] // 2):
        color = np.average(frame[i, ...])
        if color > threshold:
            low_bound_ver = i
            break

    for i in range(frame.shape[0] - 1, frame.shape[0] // 2 - 1, -1):
        color = np.average(frame[i, ...])
        if color > threshold:
            high_bound_ver = i
            break

    for j in range(0, frame.shape[1] // 2):
        color = np.average(frame[:, j, ...])
        if color > threshold:
            low_bound_hor = j
            break

    for j in range(frame.shape[1] - 1, frame.shape[1] // 2 - 1, -1):
        color = np.average(frame[:, j, ...])
        if color > threshold:
            high_bound_hor = j
            break

    return low_bound_ver, high_bound_ver + 1, low_bound_hor, high_bound_hor + 1


def write_in_info(info_row, file_name="output.csv", mode='a'):
    """
    Write a row of information to the csv file with a python list (not numpy array).
    All the values in numpy array (such as colors) are recommended to be converted to the list
    before write into the info_row.

    :param mode: The mode for using the file
    :param info_row: A row of data that will be write to the target csv file
    :param file_name: The name of the csv file
    """
    with open(file_name, newline="", mode=mode) as file:
        file_writer = csv.writer(file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        file_writer.writerow(info_row)



""" SpecifyMetaDataWindow class """

import tkinter
import copy

from kalmus.tkinter_windows.gui_utils import resource_path

# Available genre options
genres = ["Comedy", "Drama", "Western", "Science Fiction", "War", "Melodrama", "Mystery/Thriller", "Action/Adventure",
          "Musical", "Romance", "Horror", "Other:"]


class SpecifyMetaDataWindow():
    """
    SpecifyMetaDataWindow Class
    GUI window for user to specify the meta information of the selected barcode
    """
    def __init__(self, meta_data_dict, barcode=None, barcode_stacks=None):
        """
        Initialize

        :param meta_data_dict: The meta information dictionary of the barcode
        :param barcode: The barcode
        :param barcode_stacks: The dictionary that stored all the barcode on memory
        """
        self.meta_data_dict = meta_data_dict
        self.barcode = barcode
        self.barcodes_stack = barcode_stacks

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Specify Meta Data")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Set up the label prompt for the film title specification
        self.title_label = tkinter.Label(self.window, text="Film Title:")
        self.title_label.grid(row=0, column=0, sticky=tkinter.W)

        # Set up the the entry for user to specify the barcode's film title
        self.title_entry = tkinter.Entry(self.window, textvariable="1", width=30)
        self.title_entry.grid(row=0, column=1, sticky=tkinter.W)

        # Set up the label prompt for the film director specification
        self.directory_label = tkinter.Label(self.window, text="Directors:")
        self.directory_label.grid(row=1, column=0, sticky=tkinter.W)

        # Set up the entry user to specify the film director
        self.directory_entry = tkinter.Entry(self.window, textvariable="2", width=30)
        self.directory_entry.grid(row=1, column=1, sticky=tkinter.W)

        # Set up the label prompt for the film's country of origin specification
        self.country_origin_label = tkinter.Label(self.window, text="Country of Origin:")
        self.country_origin_label.grid(row=2, column=0, sticky=tkinter.W)

        # Set up the entry for user to specify the film's country of origin
        self.country_origin_entry = tkinter.Entry(self.window, textvariable="3", width=30)
        self.country_origin_entry.grid(row=2, column=1, sticky=tkinter.W)

        # Set up the label prompt for the produced year specification
        self.year_label = tkinter.Label(self.window, text="Produced Year:")
        self.year_label.grid(row=3, column=0, sticky=tkinter.W)

        # Set up the entry for user to specify the produced year of the film
        self.year_entry = tkinter.Entry(self.window, textvariable="4", width=30)
        self.year_entry.grid(row=3, column=1, sticky=tkinter.W)

        # Set up the label prompt for the film genre specification
        self.genre_label = tkinter.Label(self.window, text="Genre:")
        self.genre_label.grid(row=4, column=0, sticky=tkinter.W)

        # Genre variable
        self.genre_var = tkinter.StringVar(self.window)
        self.genre_var.set("Comedy")
        self.genre_var.trace("w", callback=self.update_other_entry)

        # Genre dropdown list
        self.genre_dropdown = tkinter.OptionMenu(self.window, self.genre_var, *genres)
        self.genre_dropdown.grid(row=4, column=1, sticky=tkinter.W)

        # Other genre entry for user to specify a non-specified (custom) genre
        self.other_entry = tkinter.Entry(self.window, textvariable="5", width=8, state="disabled")
        self.other_entry.grid(row=4, column=2, sticky=tkinter.W)

        # Update the meta data button
        self.update_meta_button = tkinter.Button(master=self.window, text="Update Meta Info",
                                                 command=self.update_meta_info)
        self.update_meta_button.grid(row=5, column=1)

    def update_meta_info(self):
        """
        Update the meta information of the barcode using user input
        """
        # Get the user input
        title = self.title_entry.get()
        director = self.directory_entry.get()
        country_of_origin = self.country_origin_entry.get()
        year = self.year_entry.get()
        genre = self.genre_var.get()

        # For non empty input
        # Update the data to the corresponding key in the meta data dictionary
        if len(title) != 0:
            self.meta_data_dict["Film Title"] = copy.deepcopy(title)
        if len(director) != 0:
            self.meta_data_dict["Directors"] = copy.deepcopy(director)
        if len(country_of_origin) != 0:
            self.meta_data_dict["Country of Origin"] = copy.deepcopy(country_of_origin)
        if len(year) != 0:
            self.meta_data_dict["Produced Year"] = copy.deepcopy(year)
        if genre != "Other:":
            self.meta_data_dict["Genre"] = copy.deepcopy(genre)
        else:
            if len(self.other_entry.get()) != 0:
                self.meta_data_dict["Genre"] = copy.deepcopy(self.other_entry.get())
            else:
                self.meta_data_dict["Genre"] = copy.deepcopy("Other")

        # If the barcode is given, update the meta data dictionary to the barcode's attributes
        if self.barcode is not None:
            self.barcode.__dict__["meta_data"] = copy.deepcopy(self.meta_data_dict)

        # If the memory stack of the barcode is given, update the barcode to the memory stack
        if self.barcodes_stack is not None:
            key_barcode = str(title) + "_" + str(director) + "_" + str(country_of_origin)\
                          + "_" + str(year) + "_" + str(genre) + "_modified_meta_info"
            self.barcodes_stack[key_barcode] = copy.deepcopy(self.barcode)

        # Quit the window
        self.window.destroy()

    def update_other_entry(self, *args):
        """
        Enable the other genre entry once user select Other: in the dropdown list

        :param args: required by the tkinter callback
        """
        # Enable or disable the other genre entry based on the dropdown list selection
        if self.genre_var.get() == "Other:":
            self.other_entry.config(state="normal")
        else:
            self.other_entry.config(state="disabled")



""" KALMUS tkinter GUI utility """

import cv2
import kalmus.utils.measure_utils as measure_utils
import os
import sys
import numpy as np
from skimage.color import hsv2rgb, rgb2hsv


def compare_two_barcodes(barcode_1, barcode_2):
    """
    Compare the similarity between two barcodes using NRMSE, SSIM, cross correlation, local cross correlation,
    Needleman Wunsch, and Smith Waterman alignment matching

    :param barcode_1: The input barcode 1
    :param barcode_2: The input barcode 2
    :return: The simiarity scores computed using all six metrics
    """
    # Resize two barcodes so they have the same shape
    # Alway resize the small barcode to the shape of the larger barcode
    target_barcode_1 = barcode_1.get_barcode().astype("float64")
    target_barcode_2 = barcode_2.get_barcode().astype("float64")
    if target_barcode_1.size <= target_barcode_2.size:
        target_barcode_1 = cv2.resize(target_barcode_1, dsize=target_barcode_2.shape[:2][::-1],
                                      interpolation=cv2.INTER_LINEAR)
    elif target_barcode_1.size > target_barcode_2.size:
        target_barcode_2 = cv2.resize(target_barcode_2, dsize=target_barcode_1.shape[:2][::-1],
                                      interpolation=cv2.INTER_LINEAR)

    # Get the cross correlation and local cross correlation
    cross_corre = measure_utils.cross_correlation(target_barcode_1.astype("float64"),
                                                  target_barcode_2.astype("float64"))
    loc_cross_corre = measure_utils.local_cross_correlation(target_barcode_1.astype("float64"),
                                                            target_barcode_2.astype("float64"))

    # Get the Normalized root mean squared error and and Structual similarity index measurement
    nrmse = measure_utils.nrmse_similarity(target_barcode_1.astype("float64"),
                                           target_barcode_2.astype("float64"))
    ssim = measure_utils.ssim_similarity(target_barcode_1.astype("float64"),
                                         target_barcode_2.astype("float64"))

    # Get the color/brightness of two barcodes
    if barcode_1.barcode_type == "Color":
        colors_barcode_1 = barcode_1.colors
        colors_barcode_2 = barcode_2.colors
    else:
        colors_barcode_1 = barcode_1.brightness
        colors_barcode_2 = barcode_2.brightness

    # Resample the barcodes if the barcodes have different size/length
    if colors_barcode_1.shape[0] > colors_barcode_2.shape[0]:
        colors_barcode_1 = colors_barcode_1[measure_utils.get_resample_index(len(colors_barcode_1),
                                                                             len(colors_barcode_2))]
    elif colors_barcode_1.shape[0] < colors_barcode_2.shape[0]:
        colors_barcode_2 = colors_barcode_2[measure_utils.get_resample_index(len(colors_barcode_2),
                                                                             len(colors_barcode_1))]
    # Convert the barcode to character string for alignment matching comparison
    # Convert the barcode to correct character based on its barcode type
    if barcode_1.barcode_type == "Color":
        string_1 = measure_utils.generate_hue_strings_from_color_barcode(colors_barcode_1)
        string_2 = measure_utils.generate_hue_strings_from_color_barcode(colors_barcode_2)
    else:
        string_1 = measure_utils.generate_brightness_string_from_brightness_barcode(colors_barcode_1)
        string_2 = measure_utils.generate_brightness_string_from_brightness_barcode(colors_barcode_2)

    # Get the Needleman Wunsch and Smith Waterman alignment matching score
    needleman = measure_utils.compare_needleman_wunsch(string_1, string_2, normalized=True)
    smithw = measure_utils.compare_smith_waterman(string_1, string_2, normalized=True)

    return cross_corre, loc_cross_corre, nrmse, ssim, needleman, smithw


def get_comparison_result_text(barcode_1, barcode_2):
    """
    Get the text comparison result using the compare_two_barcodes function

    :param barcode_1: The compared barcode 1
    :param barcode_2: The compared barcode 2
    :return: The text comparison result
    """
    # Return error text if two barcodes are not in the same type
    if barcode_1.barcode_type != barcode_2.barcode_type:
        result_text = "ERROR:\nComparison between different types of\n" \
                      "barcode type {:s} and {:s}\nare not allowed.".format(barcode_1.barcode_type,
                                                                            barcode_2.barcode_type)
        return result_text

    # Get the similarity score from the compare_two_barcodes function
    cross_corre, loc_cross_corre, nrmse, ssim, needleman, smith = compare_two_barcodes(barcode_1, barcode_2)

    # Get the text result
    result_text = "Comparison metrics:\n" \
                  "NRMSE Similarity:                  {:10.4f}\n" \
                  "SSIM Similarity:                     {:10.4f}\n" \
                  "Cross Correlation:                 {:10.4f}\n" \
                  "Local Cross Correlation:         {:10.4f}\n" \
                  "Needleman Wunsch Similarity:{:10.4f}\n" \
                  "Smith Waterman Similarity:     {:10.4f}".format(nrmse, ssim, cross_corre, loc_cross_corre,
                                                                   needleman, smith)
    return result_text


def update_graph(barcode_1, barcode_2, axes, bin_step=5):
    """
    Update the plotted graph (in place)

    :param barcode_1: The barcode 1
    :param barcode_2: The barcode 2
    :param axes: The axes of the plotted figure
    :param bin_step: The step of histogram bin
    """
    # Plot the barcode with the correct color map based on their barcode types
    if barcode_1.barcode_type == "Brightness":
        axes[0][0].imshow(barcode_1.get_barcode().astype("uint8"), cmap='gray', vmin=0, vmax=255)
    else:
        axes[0][0].imshow(barcode_1.get_barcode().astype("uint8"))

    if barcode_2.barcode_type == "Brightness":
        axes[1][0].imshow(barcode_2.get_barcode().astype("uint8"), cmap='gray', vmin=0, vmax=255)
    else:
        axes[1][0].imshow(barcode_2.get_barcode().astype("uint8"))

    # Rescale the axis range
    # And update the plot
    for axis in axes.ravel():
        axis.relim()
        axis.autoscale_view()

    # Update the title of the plotted axes
    update_axes_title(axes, barcode_1, barcode_2)

    # Update the histogram of the barcode 1
    update_hist(barcode_1, ax=axes[0][1], bin_step=bin_step)

    # Update the histogram of the barcode 2
    update_hist(barcode_2, ax=axes[1][1], bin_step=bin_step)


def update_axes_title(axes, barcode_1, barcode_2):
    """
    Update the title of the plotted axes (in place)

    :param axes: The plotted axes of the figure
    :param barcode_1: The barcode 1
    :param barcode_2: The barcode 2
    """
    title_1 = "Barcode 1"
    title_2 = "Barcode 2"

    keys = ["Film Title", "Produced Year", "Genre"]

    # Update the meta data into the title of the plotted figure
    if barcode_1.meta_data is not None:
        for key in keys:
            if key in barcode_1.meta_data.keys():
                title_1 += "    {:s}: {:s}".format(key, barcode_1.meta_data[key])

    # Update the Frame sampling type and color/brightness metric into the title
    title_1 += "    Frame Type:{:s}    {:s} Metric:{:s}".format(barcode_1.frame_type, barcode_1.barcode_type,
                                                                barcode_1.color_metric)

    # Update the meta data into the title of the plotted figure
    if barcode_2.meta_data is not None:
        for key in keys:
            if key in barcode_2.meta_data.keys():
                title_2 += "    {:s}: {:s}".format(key, barcode_2.meta_data[key])

    # Update the Frame sampling type and color/brightness metric into the title
    title_2 += "    Frame Type:{:s}    {:s} Metric:{:s}".format(barcode_2.frame_type, barcode_2.barcode_type,
                                                                barcode_2.color_metric)

    # Set up the title
    axes[0][0].set_title(title_1, fontsize=8)
    axes[1][0].set_title(title_2, fontsize=8)

    # Change the histogram's label and xticks accordingly
    if barcode_1.barcode_type == "Color":
        axes[0][1].set_xticks(np.arange(0, 361, 30))
        axes[0][1].set_xlabel("Color Hue (0 - 360)")
        axes[0][1].set_ylabel("Number of frames")
    elif barcode_1.barcode_type == "Brightness":
        axes[0][1].set_xticks(np.arange(0, 256, 16))
        axes[0][1].set_xlabel("Brightness (0 - 255)")
        axes[0][1].set_ylabel("Number of frames")

    if barcode_2.barcode_type == "Color":
        axes[1][1].set_xticks(np.arange(0, 361, 30))
        axes[1][1].set_xlabel("Color Hue (0 - 360)")
        axes[1][1].set_ylabel("Number of frames")
    elif barcode_2.barcode_type == "Brightness":
        axes[1][1].set_xticks(np.arange(0, 256, 32))
        axes[1][1].set_xlabel("Brightness (0 - 255)")
        axes[1][1].set_ylabel("Number of frames")


def update_hist(barcode, ax, bin_step=5):
    """
    Update the histogram of the plotted figure (in place)

    :param barcode: The barcode of the corresponding histogram
    :param ax: The axis that contain the histogram
    :param bin_step: The step of the bin in the plotted histogram
    """
    bin_step = bin_step

    # Plot the histogram based on the barcode's type
    if barcode.barcode_type == "Color":
        # If the barcode type is color
        # Then plot the barcode's hue value
        normalized_barcode = barcode.get_barcode().astype("float") / 255

        hsv_colors = rgb2hsv(normalized_barcode.reshape(-1, 1, 3))
        hue = hsv_colors[..., 0] * 360

        N, bins, patches = ax.hist(hue[:, 0], bins=(np.arange(0, 361, bin_step)))

        # Paint each bin with its corresponding color in hue
        paint_hue_hist(bin_step, patches)
    else:
        # If the barcode type is brightness
        N, bins, patches = ax.hist(barcode.brightness[:, 0], bins=(np.arange(0, 256, bin_step)))

        # Then paint each bin with its brightness intensity
        paint_gray_hist(bin_step, patches, opacity=0.9)


def paint_hue_hist(bin_step, patches):
    """
    Paint each bin of the hue histogram with its corresponding color in hue (in place)

    :param bin_step: The step of the bin in the histogram
    :param patches: The patches of the histogram that will be painted later
    """
    # Paint each bin of the histogram
    for i in range(len(patches)):
        # Get the hue first
        hue = np.array([i * bin_step / 360.0, 1.0, 1.0]).astype("float64")
        hue = np.expand_dims(np.expand_dims(hue, 0), 0)

        # Then get the corresponding color
        rgb = hsv2rgb(hue).tolist()
        rgb_tuple = (rgb[0][0][0], rgb[0][0][1], rgb[0][0][2])

        # Paint each patch of the histogram
        patches[i].set_facecolor(rgb_tuple)


def paint_gray_hist(bin_step, patches, opacity=0.8):
    """
    Paint each bin of the brightness histogram with its corresponding brightness intensity (in place)

    :param bin_step: The step of the bin in the histogram
    :param patches: The patches of the histogram
    :param opacity: Opacity of the brightness intensity to avoid totally white in the high intensity bin \
                    To make the histogram bin in high intensity area more observable
    """
    for i in range(len(patches)):
        # Get the RGB gray color for each bin
        bri_rgb_tuple = (bin_step * i * opacity / 255, bin_step * i * opacity / 255, bin_step * i * opacity / 255)
        patches[i].set_facecolor(bri_rgb_tuple)


def resource_path(relative_path):
    """
    Internal utility function
    Use to convert the input relative path to the absolute path
    This is used for the Pyinstaller wrapper

    :param relative_path: The relative path to the file
    :return: The absolute path to the file
    """
    try:
        # If the MEIPASS base path exist
        base_path = sys._MEIPASS
    except Exception:
        # Otherwise use the absolute path
        if relative_path.endswith(".ico"):
            # If the path points to .ico image
            if os.name != "nt":
                # If the running os is linux
                relative_path = relative_path[:-3]
                # Replace the .ico image with image in .xbm format
                relative_path += "xbm"

        # Join the path
        base_path = os.path.abspath(os.path.dirname(__file__))
        base_path = os.path.join(base_path, '../data')

        if relative_path.endswith(".xbm"):
            # If the image is .xbm image
            if os.name != "nt":
                # and the running os is linux
                # Add @ identifier at the front of path
                base_path = "@" + base_path

    # Return the path
    return os.path.join(base_path, relative_path)



"""
PlotBarcodeWindow Class
ColorHistogramWindow Class
RGBColorCubeWindow Class
OutputCSVWindow Class
"""

import tkinter
import tkinter.filedialog
from tkinter.messagebox import showinfo, showerror

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, NavigationToolbar2Tk)
import numpy as np
import os

from kalmus.utils.visualization_utils import show_colors_in_cube
from kalmus.tkinter_windows.gui_utils import resource_path, update_hist
import kalmus.utils.artist
from skimage.color import rgb2hsv


class PlotBarcodeWindow():
    """
    PlotBarcodeWindow Class
    GUI window of plotting the barcode for user to inspect in details
    """
    def __init__(self, barcode, figsize=(6, 4), dpi=100):
        """
        Initialize

        :param barcode: The barcode that will be inspected
        :param figsize: The size of the plotted figure
        :param dpi: The dpi of the figure
        """
        self.barcode = barcode

        # Initialize the window
        self.plot_window = tkinter.Tk()
        self.plot_window.wm_title("Inspect Barcode")
        self.plot_window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Set up the plotted figure
        self.fig = plt.figure(figsize=figsize, dpi=dpi)

        # Use the correct color map based on the input barcode's type
        if barcode.barcode_type == "Brightness":
            plt.imshow(barcode.get_barcode().astype("uint8"), cmap="gray")
        else:
            plt.imshow(barcode.get_barcode().astype("uint8"))
        plt.axis("off")
        plt.tight_layout()

        # Set up the canvas for the figure
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.plot_window)  # A tk.DrawingArea.
        self.canvas.draw()

        # Dynamic layout based on the type of the inspected barcode
        if barcode.barcode_type == "Color":
            column_span = 3
        else:
            column_span = 2
        self.canvas.get_tk_widget().grid(row=0, column=0, columnspan=column_span, pady=3)

        # Use tkinter Frame to organize the figure widget
        toolbarFrame = tkinter.Frame(master=self.plot_window)
        toolbarFrame.grid(row=2, column=0, columnspan=column_span, sticky=tkinter.W, pady=6)

        # Initialize the plotting tool bar
        self.toolbar = NavigationToolbar2Tk(self.canvas, toolbarFrame)
        self.toolbar.update()

        # Button to output the data in the barcode to a csv file
        self.button_output_csv = tkinter.Button(master=self.plot_window, text="Output CSV",
                                                command=self.output_csv)
        self.button_output_csv.grid(row=1, column=0, padx=18)

        # Button to check the histogram distribution of the barcode's hue/brightness value
        self.button_hist = tkinter.Button(master=self.plot_window, text="Show Histogram",
                                              command=self.show_color_histogram)
        self.button_hist.grid(row=1, column=1, padx=14)

        # If the barcode is a color barcode, allow user to inspect the RGB color distribution in a RGB cube
        if barcode.barcode_type == "Color":
            self.button_cube = tkinter.Button(master=self.plot_window, text="Show Color in RGB Cube",
                                              command=self.show_RGB_color_in_cube)
            self.button_cube.grid(row=1, column=2)

    def show_RGB_color_in_cube(self):
        """
        Instantiate the RGBColorCubeWindow if user press the show color in RGB cube button
        """
        RGBColorCubeWindow(self.barcode)

    def show_color_histogram(self):
        """
        Instantiate the ColorHistogramWindow if user press the show histogram button
        """
        ColorHistogramWindow(self.barcode)

    def output_csv(self):
        OutputCSVWindow(self.barcode)


class ColorHistogramWindow():
    """
    ColorHistogramWindow Class
    GUI window that show the distribution of the barcode's hue[0, 360]/brightness[0, 255] value
    """
    def __init__(self, barcode):
        """
        Initialize

        :param barcode: The input barcode
        """
        # Set up the window
        self.window = tkinter.Tk()
        self.window.wm_title("Histogram Distribution")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Set up the plotted figure
        fig, ax = plt.subplots(figsize=(9, 5))

        update_hist(barcode, ax=ax, bin_step=5)

        # Plot the histogram based on the barcode's type
        if barcode.barcode_type == "Color":
            ax.set_xticks(np.arange(0, 361, 30))
            ax.set_xlabel("Color Hue (0 - 360)")
            ax.set_ylabel("Number of frames")
        else:
            ax.set_xticks(np.arange(0, 255, 15))
            ax.set_xlabel("Brightness (0 - 255)")
            ax.set_ylabel("Number of frames")

        # Set up the canvas of the figure
        canvas = FigureCanvasTkAgg(fig, master=self.window)  # A tk.DrawingArea.
        canvas.draw()
        canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

        # Set up the tool bar of the figure
        toolbar = NavigationToolbar2Tk(canvas, self.window)
        toolbar.update()
        canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

class RGBColorCubeWindow():
    """
    RGBColorCubeWindow Class
    GUI window that shows the distribution of the barcode's RGB color in a RGB cube
    range in [0, 255] for all three channels
    """
    def __init__(self, barcode):
        """
        Initialize

        :param barcode: The input barcode
        """
        self.barcode = barcode

        # Set up the window
        self.window = tkinter.Tk()
        self.window.wm_title("Colors in RGB cube")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Set up the plotted figure
        sampling = 6000
        if sampling > self.barcode.colors.shape[0]:
            sampling = self.barcode.colors.shape[0]
        fig, ax = show_colors_in_cube(self.barcode.colors, return_figure=True, figure_size=(6, 6), sampling=sampling)

        # Set up the canvas
        canvas = FigureCanvasTkAgg(fig, master=self.window)  # A tk.DrawingArea.
        canvas.draw()
        canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

        # Allow mouse events on 3D figure
        ax.mouse_init()

        # Set up the tool bar of the figure
        toolbar = NavigationToolbar2Tk(canvas, self.window)
        toolbar.update()
        canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)


class OutputCSVWindow():
    """
    OutputCSVWindow class
    GUI window that outputs the per frame level color/brightness data of the inspected barcode
    The data output are stored in the csv file, and the data frame depends on the type of the barcode
    """
    def __init__(self, barcode):
        """
        Initialize

        :param barcode: The barcode to output the per frame level data
        """
        self.barcode = barcode

        # Set up the window
        self.window = tkinter.Tk()
        self.window.wm_title("Output the barcode to CSV")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Label prompt for the file name/path to the csv file
        filename_label = tkinter.Label(self.window, text="CSV file path: ")
        filename_label.grid(row=0, column=0, sticky=tkinter.W)

        # Text entry for user to type the file name/path to the csv file
        self.filename_entry = tkinter.Entry(self.window, textvariable="", width=40)
        self.filename_entry.grid(row=0, column=1, columnspan=1, sticky=tkinter.W)

        # Button to browse the folder
        self.button_browse_folder = tkinter.Button(self.window, text="Browse", command=self.browse_folder)
        self.button_browse_folder.grid(row=0, column=2)

        # Button to build/load the barcode using the given json file
        self.button_build_barcode = tkinter.Button(self.window, text="Output", command=self.output_csv_file)
        self.button_build_barcode.grid(row=1, column=1, columnspan=1)

    def output_csv_file(self):
        """
        Output the per frame level data to a csv file
        """
        # Get the file name of the output csv file
        csv_filename = self.filename_entry.get()

        if len(csv_filename) == 0:
            showerror("Invalid Path or Filename", "Please specify the path/filename of the generated csv file.\n")
            return

        # Get the sampled frame rate of the barcode
        sample_rate = self.barcode.sampled_frame_rate

        # Get the starting/skipped over frame of the barcode
        starting_frame = self.barcode.skip_over

        # Generate the corresponding csv file for the type of the barcode
        if self.barcode.barcode_type == 'Color':
            # Data frame of the csv file for the color barcode
            data_frame = ['Frame index', 'Red (0-255)', 'Green (0-255)', 'Blue (0-255)', 'Hue (0 -360)',
                          'Saturation (0 - 1)', 'Value (lightness) (0 - 1)','Brightness']
            kalmus.utils.artist.write_in_info(data_frame, csv_filename, mode='w')

            cur_frame = starting_frame
            # Get the per frame level color data
            for color in self.barcode.colors:
                r = int(color[0])
                g = int(color[1])
                b = int(color[2])
                brightness = int(0.299 * r + 0.587 * g + 0.114 * b)
                hsv = rgb2hsv(color.reshape(1, 1, -1).astype("float64") / 255)[0, 0]
                h = int(hsv[0] * 360)
                s = float(hsv[1])
                v = float(hsv[2])

                kalmus.utils.artist.write_in_info([cur_frame, r, g, b, h, s, v, brightness], csv_filename)

                cur_frame += sample_rate

        elif self.barcode.barcode_type == 'Brightness':
            # Data frame of the csv file for the brightness barcode
            data_frame = ['Frame index', 'Brightness']
            kalmus.utils.artist.write_in_info(data_frame, csv_filename, mode='w')

            cur_frame = starting_frame
            # Get the per frame level brightness data
            for bri in self.barcode.brightness:
                kalmus.utils.artist.write_in_info([cur_frame, int(bri[0])], csv_filename)

                cur_frame += sample_rate

        # Quit the window after outputting csv file
        self.window.destroy()

        showinfo("CSV File Generated Successfully", "CSV file has been generated successfully.\n"
                                                    "Path to the File: {:20s}".format(os.path.abspath(csv_filename)))

    def browse_folder(self):
        """
        Browse the folder to locate the json file
        """
        # Get the file name from the user selection
        filename = tkinter.filedialog.asksaveasfilename(initialdir=".", title="Select CSV file",
                                                      filetypes=(("csv files", "*.csv"), ("txt files", "*.txt"),
                                                                 ("All files", "*.*")))

        # Update the file name to the file name text entry
        self.filename_entry.delete(0, tkinter.END)
        self.filename_entry.insert(0, filename)



"""
Barcode Class
ColorBarcode Class
BrightnessBarcode Class
"""

import copy
import json

import cv2
import numpy as np
import threading

from kalmus.utils import artist as Artist
from kalmus.utils.artist import get_letter_box_from_frames, get_contrast_matrix_and_labeled_image

# Available metrics for computing the color of a frame
color_metrics = ["Average", "Median", "Mode", "Top-dominant", "Weighted-dominant",
                 "Brightest", "Bright"]


def foreback_segmentation(frame):
    """
    Helper function
    Segmented the input frame into two parts: foreground and background, using the GrabCut

    :param frame: Input frame
    :return: 1D image of the foreground part of the image, and 1D image of the background part of the image \
             Expected shape== Number of pixels x channels
    """
    fore_frame, back_frame = Artist.grabcut_foreback_segmentation(frame, start_row=0, row_size=frame.shape[0] - 1,
                                                                  start_col=frame.shape[1] // 6,
                                                                  col_size=frame.shape[1] * 2 // 3)
    return fore_frame, back_frame


class Barcode:
    """
    Barcode Class
    """

    def __init__(self, color_metric, frame_type, sampled_frame_rate=1, skip_over=0, total_frames=10,
                 barcode_type=None):
        """
        Initialize the barcode with the given parameters

        :param color_metric: The metric for computing the color of the frame
        :param frame_type: The type of frame sampling
        :param sampled_frame_rate: Frame sample rate: the frame sampled from every sampled_frame_rate.
        :param skip_over: The number of frames to skip with at the beginning of the video
        :param total_frames: The total number of frames (computed) included in the barcode
        :param barcode_type: The type of the barcode
        """
        self.color_metric = color_metric
        self.frame_type = frame_type
        self.barcode_type = barcode_type
        self.meta_data = {}

        self.sampled_frame_rate = sampled_frame_rate
        self.skip_over = skip_over
        self.total_frames = total_frames

        self.video = None
        self.film_length_in_frames = 0

        self.fps = None
        self.scale_factor = 1

        self.user_defined_letterbox = False
        self.low_bound_ver = 0
        self.high_bound_ver = 0
        self.low_bound_hor = 0
        self.high_bound_hor = 0

        self.barcode = None

        self.save_frames_in_generation = False
        self.saved_frames = None
        self.saved_frames_sampled_rate = -1
        self.saved_frame_height = 0

        self.rescale_frames_in_generation = False
        self.rescale_frame_factor = -1

    def read_video(self, video_path_name):
        """
        Read in the video from the given path

        :param video_path_name: The path to the video file
        """
        self.video = cv2.VideoCapture(video_path_name)

        # Get the fps of the video
        self.fps = self.video.get(cv2.CAP_PROP_FPS)

        # Get the length of the video
        self.film_length_in_frames = int(self.video.get(cv2.CAP_PROP_FRAME_COUNT))
        if self.total_frames + self.skip_over > self.film_length_in_frames:
            self.total_frames = self.film_length_in_frames - self.skip_over

        # Find the letter box
        if not self.user_defined_letterbox:
            self.find_film_letterbox()

        if self.save_frames_in_generation:
            self._determine_save_frame_param()

    def find_film_letterbox(self, num_sample=30):
        """
        Automatically find the letter box bounds of the film.
        Function run the get_letter_box_from_frames helper function by num_sample times and take the median of bounds

        :param num_sample: Number of times running the get_letter_box_from_frames
        """
        possible_indexes = np.arange(self.film_length_in_frames // 6, self.film_length_in_frames * 5 // 6, 1)
        frame_indexes = np.random.choice(possible_indexes, num_sample, replace=True)

        possible_low_bound_ver = []
        possible_high_bound_ver = []
        possible_low_bound_hor = []
        possible_high_bound_hor = []

        for frame_index in frame_indexes:
            self.video.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
            success, frame = self.video.read()
            if success:
                low_bound_v, high_bound_v, low_bound_h, high_bound_h = get_letter_box_from_frames(frame)
                possible_low_bound_ver.append(low_bound_v)
                possible_high_bound_ver.append(high_bound_v)
                possible_low_bound_hor.append(low_bound_h)
                possible_high_bound_hor.append(high_bound_h)

        self.low_bound_ver = int(np.median(possible_low_bound_ver))
        self.high_bound_ver = int(np.median(possible_high_bound_ver))
        self.low_bound_hor = int(np.median(possible_low_bound_hor))
        self.high_bound_hor = int(np.median(possible_high_bound_hor))

    def remove_letter_box_from_frame(self, frame):
        """
        Remove the letter box from the frame using the known letter box bounds

        :param frame: Input original frame with letter box
        :return: Cropped frame without letter box
        """
        frame = frame[self.low_bound_ver: self.high_bound_ver, self.low_bound_hor: self.high_bound_hor, ...]
        return frame

    def process_frame(self, frame):
        """
        Process the original frame by cropping out the letter box and resample frame using the given frame type

        :param frame: Input orignal frame
        :return: The processed and sampled frame
        """
        frame = self.remove_letter_box_from_frame(frame)
        if self.rescale_frames_in_generation:
            frame = self._resize_frame(frame)

        if self.frame_type == "Whole_frame":
            processed_frame = frame
        elif self.frame_type == "High_contrast_region":
            contrast_matrix, labels = get_contrast_matrix_and_labeled_image(frame)
            highest_contrast_region = np.sum(contrast_matrix, axis=1).argmax()
            processed_frame = frame[labels == (highest_contrast_region + 1)]
        elif self.frame_type == "Low_contrast_region":
            contrast_matrix, labels = get_contrast_matrix_and_labeled_image(frame)
            lowest_contrast_region = np.sum(contrast_matrix, axis=1).argmin()
            processed_frame = frame[labels == (lowest_contrast_region + 1)]
        elif self.frame_type == "Foreground":
            fore_frame, _ = foreback_segmentation(frame)
            if fore_frame.size == 0:
                # Empty foreground part use the whole frame instead
                fore_frame = frame
            processed_frame = fore_frame
        elif self.frame_type == "Background":
            _, back_frame = foreback_segmentation(frame)
            if back_frame.size == 0:
                # Empty background part use the whole frame instead
                back_frame = frame
            processed_frame = back_frame

        return processed_frame

    def _resize_frame(self, frame):
        frame = cv2.resize(frame, dsize=(0, 0), fx=self.rescale_frame_factor, fy=self.rescale_frame_factor,
                           interpolation=cv2.INTER_NEAREST)
        return frame

    def get_color_from_frame(self, frame):
        """
        Compute the color of the input frame using the known color metric

        :param frame: Input frame
        :return: The color of the frame computed using the known color metric
        """
        if self.color_metric == "Average":
            color = Artist.compute_mean_color(frame)
        elif self.color_metric == "Median":
            color = Artist.compute_median_color(frame)
        elif self.color_metric == "Mode":
            color, count = Artist.compute_mode_color(frame)
        elif self.color_metric == "Top-dominant":
            colors, dominances = Artist.compute_dominant_color(frame, n_clusters=3)
            pos = np.argsort(dominances)[-1]
            color = colors[pos]
        elif self.color_metric == "Weighted-dominant":
            colors, dominances = Artist.compute_dominant_color(frame, n_clusters=3)
            color = np.sum(colors * dominances.reshape(dominances.shape[0], 1), axis=0)
        elif self.color_metric == "Brightest":
            grey_frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
            pos = np.argwhere(grey_frame == grey_frame.max())[0]
            color = frame[pos[0], pos[1]].copy().astype("uint8")
        elif self.color_metric == "Bright":
            labels, bright_locations, dominance = Artist.find_bright_spots(frame, n_clusters=3, return_all_pos=True)
            top_bright = np.argsort(dominance)[-1]
            top_bright_pos = (labels == top_bright)[:, 0]
            pos = bright_locations[top_bright_pos]
            frame = frame[pos[:, 0], pos[:, 1]].reshape(pos.shape[0], 1, 3)
            color = Artist.compute_mean_color(frame)

        return color

    def get_barcode(self):
        """
        Return the barcode. If not exist reshape the stored computed colors/brightness first to get the barcode

        :return: Return the barcode
        """
        if self.barcode is None:
            self.reshape_barcode()
        return self.barcode

    def set_letterbox_bound(self, up_vertical_bound, down_vertical_bound,
                            left_horizontal_bound, right_horizontal_bound):
        """
        Manually set up the letter box bound of the film

        :param up_vertical_bound: The lower vertical bound
        :param down_vertical_bound: The higher vertical bound
        :param left_horizontal_bound: The left vertical bound
        :param right_horizontal_bound: The right vertical bound
        """
        self.enable_user_defined_letterbox()
        self.low_bound_ver = up_vertical_bound
        self.high_bound_ver = down_vertical_bound
        self.low_bound_hor = left_horizontal_bound
        self.high_bound_hor = right_horizontal_bound

    def enable_user_defined_letterbox(self):
        """
        Use the user defined letter box
        """
        self.user_defined_letterbox = True

    def automatic_find_letterbox(self):
        """
        Automatically find the letter box
        """
        self.find_film_letterbox()

    def enable_rescale_frames_in_generation(self, rescale_factor=1):
        assert rescale_factor > 0, "Rescale factor must be Positive"
        self.rescale_frames_in_generation = True
        self.rescale_frame_factor = np.sqrt(rescale_factor)

    def add_meta_data(self, key, value):
        """
        Add the meta information that describes the barcode

        :param key: The key for the meta information
        :param value: The value stored in that key
        """
        assert key is not None, "The key for the adding data cannot be None"
        if self.meta_data is None:
            self.meta_data = {}
        self.meta_data[key] = value

    def enable_save_frames(self, sampled_rate=4):
        """
        Set the save frame in the generation of barcode to be True.
        This attribute, saved_frames_in_generation, should only be modified before the generation of barcode.
        Once the barcode is generated, this attribute should not be changed.

        :param sampled_rate: Save 1 frame every sampled_rate seconds
        """
        self.save_frames_in_generation = True
        self.saved_frames = []

        self.saved_frames_sampled_rate = sampled_rate

    def _determine_save_frame_param(self):
        """
        Private method
        Determine the parameters of saving frame during the generation process.
        At most 900 Frames will be saved for each barcode
        Save frame rate is, by default, saving one frame every 4 seconds
        Frame will resized to the width of 100 pixels with the same aspect ratio
        """
        assert self.video is not None, "Video must be read before determining the save frame rate"
        assert self.fps is not None, "FPS must be determined before determining the save frame rate"
        self.saved_frames_sampled_rate = round(self.fps * self.saved_frames_sampled_rate / self.sampled_frame_rate)
        sampled_rate_upper_bound = round(self.total_frames / (self.sampled_frame_rate * 900))
        if self.saved_frames_sampled_rate < sampled_rate_upper_bound:
            self.saved_frames_sampled_rate = sampled_rate_upper_bound

        # If the barcode is too short
        if self.saved_frames_sampled_rate <= 0:
            self.saved_frames_sampled_rate = 1

        height = self.high_bound_ver - self.low_bound_ver
        width = self.high_bound_hor - self.low_bound_hor

        aspect_ratio = height / width
        self.saved_frame_height = int(100 * aspect_ratio)

    def save_frames(self, cur_used_frame, frame, frame_arr=None):
        """
        Private method
        Save the frame during the generation process.
        This functions should only be invoked during the generation process.

        :param cur_used_frame: How many frames have been read in
        :param frame: Current frame (original unprocessed frame)
        :param frame_arr: Array that stored the saved frames
        """
        if cur_used_frame % self.saved_frames_sampled_rate == 0:
            frame = self.remove_letter_box_from_frame(frame)
            resized_frame = cv2.resize(frame, dsize=(100, self.saved_frame_height))
            if frame_arr is not None:
                frame_arr.append(resized_frame)
            else:
                self.saved_frames.append(resized_frame)

    def save_as_json(self, filename=None):
        """
        Save the barcode into the json file

        :param filename: The name of the saved json file
        """
        if self.barcode is None:
            self.reshape_barcode()
        # This cv2 captured video is not pickled in the json
        # therefore it is not able to be pickled for deepcopy
        # Delete it from the object first
        self.video = None
        barcode_dict = copy.deepcopy(self.__dict__)
        barcode_dict['barcode'] = barcode_dict['barcode'].tolist()
        if self.save_frames_in_generation:
            barcode_dict['saved_frames'] = barcode_dict['saved_frames'].tolist()
        if isinstance(self, ColorBarcode):
            barcode_dict['colors'] = barcode_dict['colors'].tolist()
            barcode_type = "Color"
        elif isinstance(self, BrightnessBarcode):
            barcode_dict['brightness'] = barcode_dict['brightness'].tolist()
            barcode_type = 'Brightness'
        barcode_dict["video"] = None

        if filename is None:
            filename = "saved_{:s}_barcode_{:s}_{:s}.json" \
                .format(barcode_type, self.frame_type, self.color_metric)

        with open(filename, "w") as file:
            json.dump(barcode_dict, file)
        file.close()


class ColorBarcode(Barcode):
    """
    Color barcode
    """

    def __init__(self, color_metric, frame_type, sampled_frame_rate=1, skip_over=0, total_frames=10,
                 barcode_type="Color"):
        """
        Initialize the barcode with the given parameters

        :param color_metric: The metric for computing the color of the frame
        :param frame_type: The type of frame sampling
        :param sampled_frame_rate: Frame sample rate: the frame sampled from every sampled_frame_rate.
        :param skip_over: The number of frames to skip with at the beginning of the video
        :param total_frames: The total number of frames (computed) included in the barcode
        :param barcode_type: The type of the barcode
        """
        super().__init__(color_metric, frame_type, sampled_frame_rate, skip_over, total_frames, barcode_type)
        self.colors = None

    def collect_colors(self, video_path_name):
        """
        Collect the colors of frames from the video

        :param video_path_name: The path to the video file
        """
        self.read_video(video_path_name)

        success, frame = self.video.read()

        used_frames = 0
        cur_frame_idx = 0 + self.skip_over

        self.video.set(cv2.CAP_PROP_POS_FRAMES, cur_frame_idx)

        colors_sequence = []

        while success and used_frames < self.total_frames and cur_frame_idx < self.film_length_in_frames:
            if (cur_frame_idx % self.sampled_frame_rate) == 0:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                if self.save_frames_in_generation:
                    self.save_frames(used_frames, frame)
                frame = self.process_frame(frame)
                if len(frame.shape) <= 2:
                    frame = frame.reshape(-1, 1, 3)

                color = self.get_color_from_frame(frame)
                colors_sequence.append(color)

                used_frames += 1

            cur_frame_idx += 1

            success, frame = self.video.read()

        self.colors = np.array(colors_sequence).astype("uint8")
        if self.save_frames_in_generation:
            self.saved_frames = np.array(self.saved_frames)

    def multi_thread_collect_colors(self, video_path_name, num_thread=4):
        """
        Collect the color of the input video using Multi-thread method

        :param video_path_name: The path to the input video
        :param num_thread: Number of threads to collect the brightness
        """
        # Correct the total frames temporarily for the multi-thread generation in order to
        # be according with the definition of total frames in single thread generation
        # where total frames == self.colors.size / 3 (channels)
        self.total_frames *= self.sampled_frame_rate
        self.read_video(video_path_name)

        thread_videos = [None] * num_thread
        thread_videos[0] = self.video
        for i in range(1, num_thread):
            thread_videos[i] = cv2.VideoCapture(video_path_name)

        threads = [None] * num_thread
        thread_results = [None] * num_thread
        step = self.total_frames // num_thread

        if self.save_frames_in_generation:
            saved_frame_results = [[]] * num_thread
        else:
            saved_frame_results = None

        for i, tid in zip(range(self.skip_over, self.skip_over + self.total_frames, step),
                          range(num_thread)):
            if tid == num_thread - 1:
                start_point = i
                break
            threads[tid] = threading.Thread(target=self.thread_collect_color_start_to_end,
                                            args=(thread_videos[tid], i, step, thread_results, tid,
                                                  saved_frame_results))
            threads[tid].start()

        threads[num_thread - 1] = threading.Thread(target=self.thread_collect_color_start_to_end,
                                                   args=(thread_videos[tid], start_point,
                                                         self.total_frames + self.skip_over - start_point,
                                                         thread_results, tid, saved_frame_results))
        threads[num_thread - 1].start()

        for i in range(num_thread):
            threads[i].join()

        # Now change the total frames back to the original input
        self.total_frames = int(self.total_frames / self.sampled_frame_rate)

        colors_sequence = [thread_results[0]]
        for i in range(1, num_thread):
            colors_sequence.append(thread_results[i])

        self.colors = np.concatenate(colors_sequence).astype("uint8")

        if self.save_frames_in_generation:
            for frame_arry in saved_frame_results:
                self.saved_frames += frame_arry
            self.saved_frames = np.array(self.saved_frames)

    def thread_collect_color_start_to_end(self, video, start_point, num_frames, results, tid, frame_saved=None):
        """
        Collect the colors from the video using the multi-threads

        :param video: The video object
        :param start_point: Start point for collecting the colors
        :param num_frames: The number of frames to collect
        :param results: The placeholder for saving the results
        :param tid: The id of the thread
        :param frame_saved: The placeholder for the saved frames
        """
        assert self.video is not None, "No video is read in to the barcode for analysis."
        cur_frame_idx = start_point
        video.set(cv2.CAP_PROP_POS_FRAMES, cur_frame_idx)

        colors_sequence = []
        frame_sequence = []

        success, frame = video.read()
        used_frames = 0
        while success and used_frames < num_frames and cur_frame_idx < (start_point + num_frames):
            if (cur_frame_idx % self.sampled_frame_rate) == 0:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                if self.save_frames_in_generation:
                    self.save_frames(used_frames, frame, frame_sequence)
                frame = self.process_frame(frame)
                if len(frame.shape) <= 2:
                    frame = frame.reshape(-1, 1, 3)

                color = self.get_color_from_frame(frame)
                colors_sequence.append(color)

                used_frames += 1

            cur_frame_idx += 1

            success, frame = video.read()

        results[tid] = colors_sequence
        if self.save_frames_in_generation:
            frame_saved[tid] = frame_sequence

    def reshape_barcode(self, frames_per_column=160):
        """
        Reshape the barcode (2 dimensional with 3 channels)

        :param frames_per_column: Number of frames per column in the reshaped barcode
        """
        if len(self.colors) % frames_per_column == 0:
            self.barcode = self.colors.reshape(frames_per_column, -1, self.colors.shape[-1], order='F')
        elif len(self.colors) < frames_per_column:
            self.barcode = self.colors.reshape(-1, 1, self.colors.shape[-1], order='F')
        else:
            truncate_bound = int(len(self.colors) / frames_per_column) * frames_per_column
            self.barcode = self.colors[:truncate_bound].reshape(frames_per_column, -1,
                                                                self.colors.shape[-1], order='F')


class BrightnessBarcode(Barcode):
    """
    Brightness Barcode Class
    """

    def __init__(self, color_metric, frame_type, sampled_frame_rate=1, skip_over=0, total_frames=10,
                 barcode_type="Brightness"):
        """
        Initialize the barcode with the given parameters

        :param color_metric: The metric for computing the color of the frame
        :param frame_type: The type of frame sampling
        :param sampled_frame_rate: Frame sample rate: the frame sampled from every sampled_frame_rate.
        :param skip_over: The number of frames to skip with at the beginning of the video
        :param total_frames: The total number of frames (computed) included in the barcode
        :param barcode_type: The type of the barcode
        """
        super().__init__(color_metric, frame_type, sampled_frame_rate, skip_over, total_frames, barcode_type)
        self.brightness = None

    def collect_brightness(self, video_path_name):
        """
        Collect the brightness from the input video

        :param video_path_name: The path to the video
        """
        self.read_video(video_path_name)

        success, frame = self.video.read()

        used_frames = 0
        cur_frame_idx = 0 + self.skip_over

        self.video.set(cv2.CAP_PROP_POS_FRAMES, cur_frame_idx)

        brightness_sequence = []

        while success and used_frames < self.total_frames and cur_frame_idx < self.film_length_in_frames:
            if (cur_frame_idx % self.sampled_frame_rate) == 0:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                if self.save_frames_in_generation:
                    self.save_frames(used_frames, frame)
                frame = self.process_frame(frame)
                if len(frame.shape) <= 2:
                    frame = frame.reshape(-1, 1, 3)

                if self.color_metric in color_metrics[-2:]:
                    color = self.get_color_from_frame(frame)
                    brightness = np.sum(color * np.array([0.299, 0.587, 0.114], dtype="float64"))
                    brightness = brightness.astype("uint8")
                    brightness_sequence.append(brightness)
                else:
                    frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
                    brightness = self.get_color_from_frame(frame)
                    brightness_sequence.append(brightness)

                used_frames += 1

            cur_frame_idx += 1

            success, frame = self.video.read()

        self.brightness = np.array(brightness_sequence).astype("uint8")
        if self.save_frames_in_generation:
            self.saved_frames = np.array(self.saved_frames)

    def multi_thread_collect_brightness(self, video_path_name, num_thread=4):
        """
        Collect the brightness of the input video using Multi-thread method

        :param video_path_name: The path to the input video
        :param num_thread: Number of threads to collect the brightness
        """
        # Correct the total frames temporarily for the multi-thread generation in order to
        # be according with the definition of total frames in single thread generation
        # where total frames == self.colors.size / 3 (channels)
        self.total_frames *= self.sampled_frame_rate
        self.read_video(video_path_name)

        thread_videos = [None] * num_thread
        thread_videos[0] = self.video
        for i in range(1, num_thread):
            thread_videos[i] = cv2.VideoCapture(video_path_name)

        threads = [None] * num_thread
        thread_results = [None] * num_thread
        step = self.total_frames // num_thread

        if self.save_frames_in_generation:
            saved_frame_results = [[]] * num_thread
        else:
            saved_frame_results = None

        for i, tid in zip(range(self.skip_over, self.skip_over + self.total_frames, step),
                          range(num_thread)):
            if tid == num_thread - 1:
                start_point = i
                break
            threads[tid] = threading.Thread(target=self.thread_collect_brightness_start_to_end,
                                            args=(thread_videos[tid], i, step, thread_results,
                                                  tid, saved_frame_results))
            threads[tid].start()

        threads[num_thread - 1] = threading.Thread(target=self.thread_collect_brightness_start_to_end,
                                                   args=(thread_videos[tid], start_point,
                                                         self.total_frames + self.skip_over - start_point,
                                                         thread_results, tid, saved_frame_results))
        threads[num_thread - 1].start()

        for i in range(num_thread):
            threads[i].join()

        # Now change the total frames back to the original input
        self.total_frames = int(self.total_frames / self.sampled_frame_rate)

        brightness_sequence = [thread_results[0]]
        for i in range(1, num_thread):
            brightness_sequence.append(thread_results[i])

        self.brightness = np.concatenate(brightness_sequence).astype("uint8")

        if self.save_frames_in_generation:
            for frame_arry in saved_frame_results:
                self.saved_frames += frame_arry
            self.saved_frames = np.array(self.saved_frames)

    def thread_collect_brightness_start_to_end(self, video, start_point, num_frames, results, tid, frame_saved=None):
        """
        Collect the brightness from the video using the multi-threads

        :param video: The video object
        :param start_point: Start point for collecting the colors
        :param num_frames: The number of frames to collect
        :param results: The placeholder for saving the results
        :param tid: The id of the thread
        :param frame_saved: The placeholder for the saved frames
        """
        assert self.video is not None, "No video is read in to the barcode for analysis."
        cur_frame_idx = start_point
        video.set(cv2.CAP_PROP_POS_FRAMES, cur_frame_idx)

        brightness_sequence = []
        frame_sequence = []

        success, frame = video.read()
        used_frames = 0
        while success and used_frames < num_frames and cur_frame_idx < (start_point + num_frames):
            if (cur_frame_idx % self.sampled_frame_rate) == 0:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                if self.save_frames_in_generation:
                    self.save_frames(used_frames, frame, frame_sequence)
                frame = self.process_frame(frame)
                if len(frame.shape) <= 2:
                    frame = frame.reshape(-1, 1, 3)

                if self.color_metric in color_metrics[-2:]:
                    color = self.get_color_from_frame(frame)
                    brightness = np.sum(color * np.array([0.299, 0.587, 0.114], dtype="float64"))
                    brightness = brightness.astype("uint8")
                    brightness_sequence.append(brightness)
                else:
                    frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
                    brightness = self.get_color_from_frame(frame)
                    brightness_sequence.append(brightness)

                used_frames += 1

            cur_frame_idx += 1

            success, frame = video.read()

        results[tid] = brightness_sequence
        if self.save_frames_in_generation:
            frame_saved[tid] = frame_sequence

    def reshape_barcode(self, frames_per_column=160):
        """
        Reshape the brightness barcode (2 dimensional with 1 channel)

        :param frames_per_column: Number of frames per column in the reshaped barcode
        """
        if len(self.brightness) % frames_per_column == 0:
            self.barcode = self.brightness.reshape(frames_per_column, -1, order='F')
        elif len(self.brightness) < frames_per_column:
            self.barcode = self.brightness.reshape(-1, 1, order='F')
        else:
            truncate_bound = int(len(self.brightness) / frames_per_column) * frames_per_column
            self.barcode = self.brightness[:truncate_bound].reshape(frames_per_column, -1, order='F')



""" DisplaySavedFramesWindow Class """

import tkinter

from kalmus.tkinter_windows.gui_utils import resource_path
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, NavigationToolbar2Tk)
import numpy as np


class DisplaySavedFramesWindow():
    """
    DisplaySavedFramesWindow Class
    GUI window for displaying saved frame of a barcode around the clicked point
    """
    def __init__(self, barcode, mouse_x, mouse_y, figsize=(10, 2), dpi=100):
        self.barcode = barcode

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Display frames")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Set up the plotted figure
        self.fig = plt.figure(figsize=figsize, dpi=dpi)

        displayed_image = self.get_frames_image_for_display(mouse_x, mouse_y)
        plt.imshow(displayed_image)
        plt.axis("off")
        plt.tight_layout()

        # Set up the canvas for the figure
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.window)  # A tk.DrawingArea.
        self.canvas.draw()
        self.canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

        # Initialize the plotting tool bar
        self.toolbar = NavigationToolbar2Tk(self.canvas, self.window)
        self.toolbar.update()
        self.canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

    def get_frames_image_for_display(self, mouse_x, mouse_y):
        """
        Get the frames around the clicked point
        :param mouse_x: The x position of the clicked point
        :param mouse_y: The y position of the clicked point
        :return: The combined sampled frames for displaying
        """
        barcode_shape = self.barcode.get_barcode().shape
        # Get the middle position of the saved frame
        cur_pos = (mouse_x * barcode_shape[0] + mouse_y) / (barcode_shape[0] * barcode_shape[1])
        frame_pos = round(cur_pos * len(self.barcode.saved_frames))

        # Get another four frames around the middle frame
        # Make sure the frame positions/indexes are valid
        if frame_pos < 2:
            frame_pos = 2
        if frame_pos > len(self.barcode.saved_frames) - 3:
            frame_pos = len(self.barcode.saved_frames) - 3
        frames = self.barcode.saved_frames[frame_pos - 2: frame_pos + 3]

        # Get the combined five frames image
        combine_image = frames[0]
        for frame in frames[1:]:
            # Combine the frames into one image
            combine_image = np.concatenate((combine_image, frame), axis=1)

        return combine_image




""" SaveBarcodeWindow Class """

import tkinter
import tkinter.filedialog
from tkinter.messagebox import showwarning, showinfo
import os

from kalmus.tkinter_windows.gui_utils import resource_path


class SaveBarcodeWindow():
    """
    SaveBarcodeWindow Class
    GUI window for user to select the barcode to save into a json file that can be later loaded back to kalmus
    """
    def __init__(self, barcode_stack):
        """
        Initialize

        :param barcode_stack: the dictionary that stored all the barcodes on the memory
        """
        self.barcode_stack = barcode_stack

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))
        self.window.wm_title("Save Barcode from Memory Stack")

        # List box that lists all the barcodes stored on the memory (shows the barcode's key in the dictionary)
        self.listbox = tkinter.Listbox(self.window, selectmode=tkinter.SINGLE, width=65, height=20)
        self.listbox.grid(row=0, column=0, columnspan=4)

        # List all the barcodes in the list box using their keys
        for barcode_names in self.barcode_stack.keys():
            self.listbox.insert(tkinter.END, barcode_names)

        # Label prompt for the file name/path to the saved json file
        filename_label = tkinter.Label(self.window, text="JSON file path: ")
        filename_label.grid(row=1, column=0)

        # Text entry for user to specify the file name/path to the saved json file
        self.filename_entry = tkinter.Entry(self.window, textvariable="", width=40)
        self.filename_entry.grid(row=1, column=1, columnspan=1, sticky=tkinter.W)

        # Button to browse the location in a file manager
        self.button_browse_folder = tkinter.Button(self.window, text="Browse", command=self.browse_folder)
        self.button_browse_folder.grid(row=1, column=2, sticky=tkinter.W)

        # Button to save the barcode into json file
        self.button_save = tkinter.Button(master=self.window, text="Save Barcode", command=self.save_stack)
        self.button_save.grid(row=2, column=0, sticky=tkinter.W)

    def browse_folder(self):
        """
        Browse the folders in a file manager window
        """
        # Get the file name/path from the user input in the file manager
        filename = tkinter.filedialog.asksaveasfilename(initialdir=".", title="Save JSON file",
                                                    filetypes=(("json files", "*.json"), ("txt files", "*.txt"),
                                                               ("All files", "*.*")))

        # Update the file name/path to the file name entry
        self.filename_entry.delete(0, tkinter.END)
        self.filename_entry.insert(0, filename)

    def save_stack(self):
        """
        Save the selected barcode from the memory stack to the json file
        """
        # Get the dictionary key of the selected barcode
        selected_barcode_names = [self.listbox.get(idx) for idx in self.listbox.curselection()]

        # Get the file name/path to the saved json file from the user input
        filename = self.filename_entry.get()

        if len(filename) == 0:
            showwarning("Default Saved JSON Path is Used", "Path to the saved JSON file is not specified.\n"
                                                           "Default save path is used.\n"
                                                           "File will be saved in the current working directory.\n"
                                                           "It is recommended to specify the file path.")
            filename = None

        # Saved the barcode
        for barcode_name in selected_barcode_names:
            barcode = self.barcode_stack[barcode_name]
            barcode.save_as_json(filename)

        # Quit the window
        self.window.destroy()

        # Show success message
        showinfo("JSON File Saved Successfully", "The JSON file is successfully saved.\n\n"
                                                 "The Path to the File: {:20s}".format(os.path.abspath(filename)))



""" Command line startup of the kalmus software """

from kalmus.barcodes.BarcodeGenerator import BarcodeGenerator
from kalmus.tkinter_windows.MainWindowVersion2 import MainWindow
import os


def main():
    # Instantiate the barcode generator object
    barcode_gn = BarcodeGenerator()
    # Build the default barcode from the default json file
    json_path = os.path.abspath(os.path.dirname(__file__))
    json_path = os.path.join(json_path, "data/mission_impossible_Bright_Whole_frame_Color.json")
    barcode_gn.generate_barcode_from_json(json_file_path=json_path,
                                          barcode_type="Color")

    # Get the default barcode
    barcode_tmp = barcode_gn.get_barcode()

    # Use the default barcode and the barcode generator to Instantiate the Main window of the kalmus software (GUI)
    MainWindow(barcode_tmp, barcode_gn)



""" CheckTimePointWindow Class """

import tkinter
from tkinter.messagebox import showerror

from kalmus.tkinter_windows.gui_utils import resource_path
from kalmus.tkinter_windows.time_points_windows.CalibrateBarcodeTimeWindow import CalibrateBarcodeTimeWindow
from kalmus.tkinter_windows.time_points_windows.DisplaySavedFramesWindow import DisplaySavedFramesWindow


class CheckTimePointWindow():
    """
    CheckTimePointWindow
    GUI window that shows the RGB/Brightness value, RGB color, x, y position, frame indexes and time
    at the clicked point
    """

    def __init__(self, barcode, mouse_x, mouse_y):
        """
        Initialize

        :param barcode: The barcode with clicked point
        :param mouse_x: The x position of the clicked point
        :param mouse_y: The y position of the clicked point
        """
        self.barcode = barcode
        self.x_pos = mouse_x
        self.y_pos = mouse_y

        # Compute the frame label
        frame = (self.barcode.skip_over + self.barcode.sampled_frame_rate * \
                ((self.x_pos * self.barcode.get_barcode().shape[0]) + self.y_pos)) * self.barcode.scale_factor
        frame = int(frame)

        # If frame rate is not given, use 30 as default
        if self.barcode.fps is None:
            self.barcode.fps = 30

        # Compute the time label
        time_tot_sec = frame / self.barcode.fps
        time_sec = int(time_tot_sec % 60)
        time_min = int((time_tot_sec / 60) % 60)
        time_hr = int(time_tot_sec / 3600)

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("At this point...")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # If the clicked barcode is color barcode
        if self.barcode.barcode_type == "Color":
            # Get the r, g, b value at that point
            r, g, b = barcode.get_barcode().astype("uint8")[self.y_pos, self.x_pos]
            # Set up the label
            color_r_label = tkinter.Label(master=self.window, text="Red: {:d}".format(r), fg='#9E1A1A')
            color_r_label.grid(row=0, column=0)

            color_g_label = tkinter.Label(master=self.window, text="Green: {:d}".format(g), fg='#028A0F')
            color_g_label.grid(row=0, column=1)

            color_b_label = tkinter.Label(master=self.window, text="Blue: {:d}".format(b), fg='#1520A6')
            color_b_label.grid(row=0, column=2)

            # Show the color
            color_label = tkinter.Label(master=self.window, text="", bg=f'#{r:02x}{g:02x}{b:02x}', width=6, height=2)
            color_label.grid(row=0, column=3, rowspan=2, padx=5, pady=5)
        elif self.barcode.barcode_type == "Brightness":
            # Get the brightness value (notices that r = g = b = brightness)
            r = g = b = barcode.get_barcode().astype("uint8")[self.y_pos, self.x_pos]
            brightness_value_label = tkinter.Label(master=self.window, text="Brightness: {:d}".format(r))
            brightness_value_label.grid(row=0, column=0, sticky=tkinter.E)

            # Set up the label
            brightness_label = tkinter.Label(master=self.window, text="", bg=f'#{r:02x}{g:02x}{b:02x}',
                                             width=12, height=1)
            # Show the brightness
            brightness_label.grid(row=0, column=1, columnspan=2, padx=5, pady=3)

        # Show the position
        pos_label = tkinter.Label(master=self.window, text="Position: ({:d}, {:d}) ".format(self.x_pos, self.y_pos))
        pos_label.grid(row=1, column=0)

        # Show the frame index
        self.frame_label = tkinter.Label(master=self.window, text="Frame: {:d} ".format(frame))
        self.frame_label.grid(row=1, column=1)

        # Show the time
        self.time_label = tkinter.Label(master=self.window, text="Time: {:02d}:{:02d}:{:02d} "
                                        .format(time_hr, time_min, time_sec))
        self.time_label.grid(row=1, column=2)

        # Calibrate barcode button
        self.cali_button = tkinter.Button(master=self.window, text="Calibrate", command=self.calibrate_barcode_time)
        self.cali_button.grid(row=2, column=1)

        self.display_button = tkinter.Button(master=self.window, text="Display", command=self.display)
        self.display_button.grid(row=2, column=2)

    def calibrate_barcode_time(self):
        """
        Instantiate the CalibrateBarcodeWindow
        """
        CalibrateBarcodeTimeWindow(self.barcode, self.time_label, self.frame_label, self.x_pos, self.y_pos)

    def display(self):
        """
        Instantiate the DisplaySavedFramesWindow if barcode has the saved frames
        """
        # Check if barcode has the saved frames
        if self.barcode.saved_frames is None:
            # If barcode does not have the saved frames
            # Show the warning
            showerror("No Saved Frames", "This Barcode does not have\nthe saved frames to display.")
        else:
            DisplaySavedFramesWindow(barcode=self.barcode, mouse_x=self.x_pos, mouse_y=self.y_pos)



""" SaveImageWindow Class """

import tkinter
from tkinter.messagebox import showerror, showinfo
import tkinter.filedialog
import matplotlib.pyplot as plt
import cv2
import os

from kalmus.tkinter_windows.gui_utils import resource_path


class SaveImageWindow():
    """
    SaveImageWindow Class
    Save the barcode into the image
    """
    def __init__(self, barcode_1, barcode_2):
        """
        Initialize
        :param barcode_1: The barcode 1
        :param barcode_2: The barcode 2
        """
        # Initialize the window
        self.window = tkinter.Tk()
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))
        self.window.wm_title("Save Image")

        self.barcode_1 = barcode_1
        self.barcode_2 = barcode_2

        # Label prompt for which barcode to save
        which_barcode_label = tkinter.Label(self.window, text="Barcode: ")
        which_barcode_label.grid(row=0, column=0, columnspan=1)

        # Barcode option variable
        self.barcode_option = tkinter.StringVar(self.window)
        self.barcode_option.set("Barcode 1")

        # Radio button for which barcode to save
        radio_barcode_1 = tkinter.Radiobutton(self.window, text="Barcode 1", variable=self.barcode_option,
                                              value="Barcode 1", command=self.update_size_entry)
        radio_barcode_1.grid(row=1, column=0)
        radio_barcode_1.select()

        radio_barcode_2 = tkinter.Radiobutton(self.window, text="Barcode 2", variable=self.barcode_option,
                                              value="Barcode 2", command=self.update_size_entry)
        radio_barcode_2.grid(row=2, column=0)

        # The width and height (in pixels) of the selected barcode
        width = self.barcode_1.get_barcode().shape[1]
        height = self.barcode_1.get_barcode().shape[0]

        # Resize the barcode into desirable size before saving
        self.resize_x_label = tkinter.Label(self.window, text="Saved Width (pixels): ")
        self.resize_x_label.grid(row=1, column=1, sticky=tkinter.E)

        self.resize_x_entry = tkinter.Entry(self.window, textvariable=-2, width=5)
        self.resize_x_entry.grid(row=1, column=2, padx=15, sticky=tkinter.W)
        self.resize_x_entry.insert(0, str(width))

        self.resize_y_label = tkinter.Label(self.window, text="Saved Height (pixels): ")
        self.resize_y_label.grid(row=2, column=1, sticky=tkinter.E)

        self.resize_y_entry = tkinter.Entry(self.window, textvariable=-3, width=5)
        self.resize_y_entry.grid(row=2, column=2, padx=15, sticky=tkinter.W)
        self.resize_y_entry.insert(0, str(height))

        # Label prompt for the file name (path) of the saved image
        filename_label = tkinter.Label(self.window, text="Image file path: ")
        filename_label.grid(row=3, column=0)

        # Text entry for user to specify the path of the saved image
        self.filename_entry = tkinter.Entry(self.window, textvariable="", width=40)
        self.filename_entry.grid(row=3, column=1, columnspan=1, sticky=tkinter.W)

        # Button to browse the location in a file manager window
        self.button_browse_folder = tkinter.Button(self.window, text="Browse", command=self.browse_folder)
        self.button_browse_folder.grid(row=3, column=2, sticky=tkinter.W)

        # Button to save the image into the given path using the given size
        self.button_save_image = tkinter.Button(master=self.window, text="Save Barcode", command=self.save_image)
        self.button_save_image.grid(row=4, column=0)

    def browse_folder(self):
        """
        Browse the folders in a file manager window
        """
        # Get the file name/path from the user input in the file manager
        filename = tkinter.filedialog.asksaveasfilename(initialdir=".", title="Save Image file",
                                                    filetypes=(("JPEG files", "*.jpg"), ("PNG files", "*.png"),
                                                               ("All files", "*.*")))

        # Update the file name/path to the file name entry
        self.filename_entry.delete(0, tkinter.END)
        self.filename_entry.insert(0, filename)

    def update_size_entry(self):
        """
        Update the size of current selected barcodes displayed in the resize entries
        """
        # Find the current selected barcode
        # Update the width and height (in pixels) of that barcode in the resize entries
        if self.barcode_option.get() == "Barcode 1":
            width = self.barcode_1.get_barcode().shape[1]
            height = self.barcode_1.get_barcode().shape[0]

        elif self.barcode_option.get() == "Barcode 2":
            width = self.barcode_2.get_barcode().shape[1]
            height = self.barcode_2.get_barcode().shape[0]

        self.resize_x_entry.delete(0, tkinter.END)
        self.resize_x_entry.insert(0, width)

        self.resize_y_entry.delete(0, tkinter.END)
        self.resize_y_entry.insert(0, height)

    def save_image(self):
        """
        Save the currently selected barcode into the image with the given size
        """
        # Check if the filename is given
        filename = self.filename_entry.get()
        if len(filename) == 0:
            showerror("File Name is Not Given", "Please specify the path to the saved image.")
            return

        # Get which barcode to save
        if self.barcode_option.get() == "Barcode 1":
            barcode = self.barcode_1.get_barcode().astype("uint8")
            barcode_type = self.barcode_1.barcode_type
        elif self.barcode_option.get() == "Barcode 2":
            barcode = self.barcode_2.get_barcode().astype("uint8")
            barcode_type = self.barcode_2.barcode_type

        # Resize the barcode into the desired shape (notice that the original barcode won't be affected)
        barcode = cv2.resize(barcode, dsize=(int(self.resize_x_entry.get()), int(self.resize_y_entry.get())),
                             interpolation=cv2.INTER_NEAREST)

        # Save the barcode with desirable color map based on its barcode type
        if barcode_type == "Color":
            plt.imsave(filename, barcode)
        else:
            plt.imsave(filename, barcode, cmap="gray")

        # Quit the window
        self.window.destroy()

        showinfo("Image Saved Successfully", "The image is saved successfully.\n\n"
                                             "The Path to the Image: {:20s}".format(os.path.abspath(filename)))



""" StatsInfoWindow Class """

import tkinter

from kalmus.tkinter_windows.gui_utils import get_comparison_result_text, resource_path


class StatsInfoWindow():
    """
    StatsInfoWindow Class
    Show the similarity scores between two barcodes using all comparing metrics
    The score are shown in text panel
    """
    def __init__(self, barcode_1, barcode_2):
        """
        Initialize the Window

        :param barcode_1: The barcode 1
        :param barcode_2: The barcode 2
        """
        self.barcode_1 = barcode_1
        self.barcode_2 = barcode_2

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Barcodes Similarity Statistics")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Get the comparison result (in string) between two barcodes
        result_text = get_comparison_result_text(self.barcode_1, self.barcode_2)

        # Show the result in a Label (text panel)
        self.info_label = tkinter.Label(self.window, text=result_text, width=35, bg='white', anchor='w',
                                        justify=tkinter.LEFT)
        self.info_label.grid(row=0, column=0, rowspan=8)



""" ReshapeBarcodeWindow Class """

import tkinter
import copy
import cv2

from kalmus.tkinter_windows.gui_utils import update_graph, resource_path


class ReshapeBarcodeWindow():
    """
    ReshapeBarcodeWindow Class
    GUI window for user to reshape the selected barcode into the desirable shape
    """
    def __init__(self, barcode_1, barcode_2, axes, canvas):
        """
        Initialize

        :param barcode_1: The barcode 1
        :param barcode_2: The barcode 2
        :param axes: The display axes in the MainWindow of the kalmus
        :param canvas: The display canvas in the MainWindow of the kalmus
        """
        self.barcode_1 = barcode_1
        self.barcode_2 = barcode_2

        self.axes = axes
        self.canvas = canvas

        # Initialize the window
        self.window = tkinter.Tk()
        self.window.wm_title("Reshape/Resize Barcode Config")
        self.window.iconbitmap(resource_path("kalmus_icon.ico"))

        # Reshape/Resize option
        self.config_option = tkinter.StringVar(self.window)
        self.config_option.set("Reshape")  # initialize

        # Prompt for the resize parameters specification
        params_label = tkinter.Label(self.window, text="Config Params: ")
        params_label.grid(row=0, column=0, columnspan=1, sticky=tkinter.W)

        # Label (text) prompt and entry for user to specify the resize parameters
        column_length_label = tkinter.Label(self.window, text="Frames per Column: ")
        column_length_label.grid(row=1, column=0, sticky=tkinter.W)

        self.column_length_entry = tkinter.Entry(self.window, textvariable="-1", width=5)
        self.column_length_entry.grid(row=1, column=1, padx=15)

        self.resize_x_label = tkinter.Label(self.window, text="Scale Width by (ratio):    ")
        self.resize_x_label.grid(row=2, column=0, sticky=tkinter.W)

        self.resize_x_entry = tkinter.Entry(self.window, textvariable="-2", width=5, state="disabled")
        self.resize_x_entry.grid(row=2, column=1, padx=15)

        self.resize_y_label = tkinter.Label(self.window, text="Scale Height by (ratio):    ")
        self.resize_y_label.grid(row=3, column=0, sticky=tkinter.W)

        self.resize_y_entry = tkinter.Entry(self.window, textvariable="-3", width=5, state="disabled")
        self.resize_y_entry.grid(row=3, column=1, padx=15)

        # Label prompt for displaying the width and height of the currently selected barcode
        self.size_label = tkinter.Label(self.window, text="Current Width = {:d}\nCurrent Height = {:d}"
                                        .format(self.barcode_1.get_barcode().shape[1],
                                                self.barcode_1.get_barcode().shape[0]))
        self.size_label.grid(row=4, column=0, columnspan=1)

        # Button to process the resize
        self.process_button = tkinter.Button(self.window, text="Process", command=self.reshape_resize_barcode)
        self.process_button.grid(row=4, column=2, sticky=tkinter.W)

        # Label prompt for the Resize type selection
        config_label = tkinter.Label(self.window, text="Config options: ")
        config_label.grid(row=0, column=2, columnspan=1)

        # Radio button for selecting the resize type
        radio_reshape = tkinter.Radiobutton(self.window, text="Reshape", variable=self.config_option,
                                            value="Reshape", anchor='w',
                                            command=self.reshape)
        radio_reshape.grid(row=1, column=2, sticky=tkinter.W)
        radio_reshape.select()

        radio_scaling = tkinter.Radiobutton(self.window, text="Scaling", variable=self.config_option,
                                            value="Scaling", anchor='w',
                                            command=self.scale)
        radio_scaling.grid(row=2, column=2, sticky=tkinter.W)

        radio_resize = tkinter.Radiobutton(self.window, text="Resize", variable=self.config_option,
                                           value="Resize", anchor='w',
                                           command=self.resize)
        radio_resize.grid(row=3, column=2, sticky=tkinter.W)

        # Label prompt for selecting which barcode to resize
        which_barcode_label = tkinter.Label(self.window, text="Barcode: ")
        which_barcode_label.grid(row=0, column=3, columnspan=1)

        # Option variable
        self.barcode_option = tkinter.StringVar(self.window)
        self.barcode_option.set("Barcode 1")

        # Radio button for selecting which barcode to resize
        radio_barcode_1 = tkinter.Radiobutton(self.window, text="Barcode 1", variable=self.barcode_option,
                                              value="Barcode 1", command=self.update_size_label)
        radio_barcode_1.grid(row=1, column=3)
        radio_barcode_1.select()

        radio_barcode_2 = tkinter.Radiobutton(self.window, text="Barcode 2", variable=self.barcode_option,
                                              value="Barcode 2", command=self.update_size_label)
        radio_barcode_2.grid(row=2, column=3)

    def update_size_label(self):
        """
        Update the size label if the currently selected barcode is changed
        """
        if self.barcode_option.get() == "Barcode 1":
            text = "Current Width = {:d}\nCurrent Height = {:d}".format(
                self.barcode_1.get_barcode().shape[1], self.barcode_1.get_barcode().shape[0])
        elif self.barcode_option.get() == "Barcode 2":
            text = "Current Width = {:d}\nCurrent Height = {:d}".format(
                self.barcode_2.get_barcode().shape[1], self.barcode_2.get_barcode().shape[0])

        self.size_label['text'] = text

    def reshape(self):
        """
        Enable or disable the input parameters entry if the reshape radio button is selected
        """
        self.column_length_entry.config(state='normal')
        self.resize_x_entry.config(state='disabled')
        self.resize_y_entry.config(state='disabled')

    def scale(self):
        """
        Enable or disable the input parameters entry and update the corresponding text
        if the scale radio button is selected
        """
        self.resize_x_label['text'] = "Scale Width by (ratio):    "
        self.resize_y_label['text'] = "Scale Height by (ratio):    "

        self.column_length_entry.config(state='disabled')
        self.resize_x_entry.config(state='normal')
        self.resize_y_entry.config(state='normal')

    def resize(self):
        """
        Enable or disable the input parameters entry and update the corresponding text
        if the resize radio button is selected
        """
        self.resize_x_label['text'] = "Resize Width to (pixels): "
        self.resize_y_label['text'] = "Resize Height to (pixels): "

        self.column_length_entry.config(state='disabled')
        self.resize_x_entry.config(state='normal')
        self.resize_y_entry.config(state='normal')

    def reshape_resize_barcode(self):
        """
        Reshape or resize the barcode using the given parameters
        """
        # Get the reshape/resize type from the user selection
        option = self.config_option.get()

        # Get which barcode to reshape/resize
        if self.barcode_option.get() == "Barcode 1":
            barcode = self.barcode_1
        elif self.barcode_option.get() == "Barcode 2":
            barcode = self.barcode_2

        # Save the current barcode size
        old_barcode_size = barcode.get_barcode().shape[0] * barcode.get_barcode().shape[1]

        # Reshape/resize the currently selected barcode using the given type with parameters
        if option == "Reshape":
            frames_per_column_str = self.column_length_entry.get()

            # Check if the reshape parameter is given
            # If not given, return and do not process the reshape
            if len(frames_per_column_str) == 0:
                return

            frames_per_column = int(frames_per_column_str)
            barcode.reshape_barcode(frames_per_column)

            self.update_scale_factor(barcode, old_barcode_size)
            self.updated_new_barcode()
        elif option == "Resize":
            barcode_shape = barcode.get_barcode().shape
            resize_x, resize_y = self._check_resize_entry(barcode_shape[1], barcode_shape[0])
            if resize_x is None:
                return

            resized_barcode = cv2.resize(barcode.get_barcode(),
                                         dsize=(int(resize_x), int(resize_y)),
                                         interpolation=cv2.INTER_NEAREST)
            barcode.barcode = resized_barcode

            self.update_scale_factor(barcode, old_barcode_size)
            self.updated_new_barcode()
        elif option == "Scaling":
            resize_x, resize_y = self._check_resize_entry(1, 1)
            if resize_x is None:
                return

            resized_barcode = cv2.resize(barcode.get_barcode(),
                                         dsize=(0, 0),
                                         fx=float(resize_x),
                                         fy=float(resize_y),
                                         interpolation=cv2.INTER_NEAREST)
            barcode.barcode = resized_barcode

            self.update_scale_factor(barcode, old_barcode_size)
            self.updated_new_barcode()

        # Quit the window
        self.window.destroy()

    def _check_resize_entry(self, default_x, default_y):
        """
        Check if the resize parameter is given
        If one of the parameter is not given assume that dimension is unchanged
        If both are not given, return and do not process the resize.

        :param default_x: Default x dimension
        :param default_y: Default y dimension
        :return: Processed resize x and y parameters from the user input
        """
        resize_x_str = self.resize_x_entry.get()
        resize_y_str = self.resize_y_entry.get()

        if len(resize_x_str) == 0 and len(resize_y_str) == 0:
            return None, None
        if len(resize_x_str) == 0:
            resize_x = str(default_x)
        else:
            resize_x = resize_x_str
        if len(resize_y_str) == 0:
            resize_y = str(default_y)
        else:
            resize_y = resize_y_str

        return resize_x, resize_y

    def update_scale_factor(self, barcode, old_barcode_size):
        """
        Update the scale factor of the barcode

        :param barcode: The barcode to update
        :param old_barcode_size: The old size of that barcode
        """
        barcode.scale_factor *= (old_barcode_size / (barcode.get_barcode().shape[0] * barcode.get_barcode().shape[1]))

    def updated_new_barcode(self):
        """
        Update the resized/reshaped barcode to the MainWindow of the kalmus
        """
        # Clear the display axes
        self.axes[0][0].cla()
        self.axes[1][0].cla()
        self.axes[0][1].cla()
        self.axes[1][1].cla()

        # Update the displayed barcode and redraw the canvas
        if self.barcode_1.get_barcode().shape[1] > self.barcode_2.get_barcode().shape[1]:
            temp = copy.deepcopy(self.barcode_1)

            self.barcode_1.__dict__ = self.barcode_2.__dict__.copy()
            self.barcode_2.__dict__ = temp.__dict__.copy()

        update_graph(barcode_1=self.barcode_1, barcode_2=self.barcode_2, axes=self.axes)

        # Redraw the canvas
        self.canvas.draw()



"""
MainWindow Class
Version2
"""

import tkinter
from tkinter.messagebox import askokcancel, showinfo, showerror

from kalmus.tkinter_windows.GenerateBarcodeWindow import GenerateBarcodeWindow
from kalmus.tkinter_windows.SaveBarcodeWindow import SaveBarcodeWindow
from kalmus.tkinter_windows.LoadStackWindow import LoadStackWindow
from kalmus.tkinter_windows.LoadJsonWindow import LoadJsonWindow
from kalmus.tkinter_windows.ReshapeBarcodeWindow import ReshapeBarcodeWindow
from kalmus.tkinter_windows.gui_utils import paint_hue_hist, update_axes_title, resource_path
from kalmus.tkinter_windows.plot_barcodes_windows.WhichBarcodeInspectWindow import WhichBarcodeInspectWindow
from kalmus.tkinter_windows.StatsInfoWindow import StatsInfoWindow
from kalmus.tkinter_windows.SaveImageWindow import SaveImageWindow
from kalmus.tkinter_windows.meta_info_windows.WhichBarcodeCheckMeta import WhichBarcodeCheckMeta
from kalmus.tkinter_windows.time_points_windows.CheckTimePointWindow import CheckTimePointWindow

import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import (FigureCanvasTkAgg, NavigationToolbar2Tk)

from skimage.color import rgb2hsv

import numpy as np

import copy
import matplotlib

# General setup of the figure
font = {'family': 'DejaVu Sans',
        'size': 8}

matplotlib.rc('font', **font)


class MainWindow():
    """
    MainWindow Class
    The main GUI window for user to interact with the Kalmus software
    Has two displays for the barcodes and the histograms of their hue/brightness values
    Has all buttons to the subwindow of the kalmus software
    """

    def __init__(self, barcode_tmp, barcode_gn, figsize=(12, 5), dpi=100):
        """
        Initialize

        :param barcode_tmp: The temporary barcode object for software initialization
        :param barcode_gn: The barcode generator object
        :param figsize: The size of the plotted figure
        :param dpi: The dpi of the plotted figure
        """
        # Initialize the barcode memory stack
        self.barcodes_stack = {"default": copy.deepcopy(barcode_tmp)}

        # Copy over the barcodes
        self.barcode_1 = copy.deepcopy(barcode_tmp)
        self.barcode_2 = copy.deepcopy(barcode_tmp)

        # Initialize the barcode's meta data to none
        self.barcode_1.meta_data = {}
        self.barcode_2.meta_data = {}

        # Get the barcode generator
        self.barcode_gn = barcode_gn

        # Initialize the window
        self.root = tkinter.Tk()

        self.root.configure(bg='#85C1FA')
        self.root.wm_title("KALMUS Version 1.3.8a0")
        self.root.iconbitmap(resource_path("kalmus_icon.ico"))

        # Initialize the figure
        self.fig, self.ax = plt.subplots(2, 2, figsize=figsize, dpi=dpi,
                                         gridspec_kw={'width_ratios': [2.8, 1]}, sharex='col', sharey='col')

        update_axes_title(self.ax, self.barcode_1, self.barcode_2)

        # Plot the barcodes into the figure
        self.barcode_display_1 = self.ax[0][0].imshow(self.barcode_1.get_barcode().astype("uint8"))
        self.barcode_display_2 = self.ax[1][0].imshow(self.barcode_2.get_barcode().astype("uint8"))

        # Normalize the barcode before converting the color map
        normalized_barcode_1 = self.barcode_1.get_barcode().astype("float") / 255

        hsv_colors_1 = rgb2hsv(normalized_barcode_1.reshape(-1, 1, 3))
        hue_1 = hsv_colors_1[..., 0] * 360

        # The step of the histogram
        bin_step = 5
        N, bins, patches = self.ax[0][1].hist(hue_1[:, 0], bins=(np.arange(0, 361, bin_step)))

        paint_hue_hist(bin_step, patches)

        # Update the hue histogram
        self.ax[0][1].set_xticks(np.arange(0, 361, 30))
        self.ax[0][1].set_xlabel("Color Hue (0 - 360)")
        self.ax[0][1].set_ylabel("Number of frames")

        normalized_barcode_2 = self.barcode_2.get_barcode().astype("float") / 255

        hsv_colors_2 = rgb2hsv(normalized_barcode_2.reshape(-1, 1, 3))
        hue_2 = hsv_colors_2[..., 0] * 360

        N, bins, patches = self.ax[1][1].hist(hue_2[:, 0], bins=(np.arange(0, 361, bin_step)))

        # Paint the hue histogram where each bin of the histogram is in the color of the corresponding hue
        paint_hue_hist(bin_step, patches)

        self.ax[1][1].set_xticks(np.arange(0, 361, 30))
        self.ax[1][1].set_xlabel("Color Hue (0 - 360)")
        self.ax[1][1].set_ylabel("Number of frames")

        # Use tight layout
        plt.tight_layout()

        # Draw the canvas
        self.canvas = FigureCanvasTkAgg(self.fig, master=self.root)  # A tk.DrawingArea.
        self.canvas.draw()

        self.canvas.mpl_connect('button_press_event', self.time_pick)

        # Use tkinter Frame to organize the figure widget
        toolbarFrame = tkinter.Frame(master=self.root)
        toolbarFrame.grid(row=8, column=2)

        # Set up the tool bar of the plotted figure
        self.toolbar = NavigationToolbar2Tk(self.canvas, toolbarFrame)
        self.toolbar.update()

        # Position the canvas/plotted figure into the window
        self.canvas.get_tk_widget().grid(row=0, column=1, rowspan=8, columnspan=3)

        self.generate_window_opened = False

        # Button to generate the barcode
        button_generate = tkinter.Button(master=self.root, text="Generate Barcode",
                                         command=self.generate_barcode)
        button_generate.grid(row=0, column=0, padx=3)

        # Button to load the barcode from existed json files
        button_load = tkinter.Button(master=self.root, text="Load JSON",
                                     command=self.load_json_barcode)
        button_load.grid(row=1, column=0)

        # Button to load the barcode from the memory stack
        button_load_stack = tkinter.Button(master=self.root, text="Load Memory",
                                           command=self.load_stack_barcode)
        button_load_stack.grid(row=2, column=0)

        # Button to reshape the barcode displayed in the main window
        button_reshape_barcode = tkinter.Button(master=self.root, text="Reshape Barcode",
                                                command=self.reshape_barcode)
        button_reshape_barcode.grid(row=3, column=0)

        # Button to save the barcode into json files
        button_save_json = tkinter.Button(master=self.root, text="Save JSON",
                                          command=self.save_barcode_on_stack)
        button_save_json.grid(row=4, column=0)

        # Button to save the barcode displayed into the image
        button_save_image = tkinter.Button(master=self.root, text="Save Image",
                                           command=self.save_image_from_display)
        button_save_image.grid(row=5, column=0)

        # Button to inspect the barcode in details
        button_barcode = tkinter.Button(master=self.root, text="Inspect Barcode",
                                        command=self.show_barcode)
        button_barcode.grid(row=6, column=0)

        # Button to show the statistics of the barcode
        button_stats_info = tkinter.Button(master=self.root, text="Stats Info",
                                           command=self.stats_info)
        button_stats_info.grid(row=7, column=0)

        # Button to quit the main window
        button_quit = tkinter.Button(master=self.root, text="Quit", command=self.close_window)
        button_quit.grid(row=8, column=0)

        # Button to check the meta data of the displayed barcodes
        button_check_meta = tkinter.Button(master=self.root, text="Check Meta Info", command=self.check_meta_info)
        button_check_meta.grid(row=8, column=3)

        # Close the window mainloop if user try to close the window
        self.root.protocol("WM_DELETE_WINDOW", self.close_window)

        # Start the main window
        self.root.mainloop()

    def close_window(self):
        """
        close the Mainwindow
        Check if the Generate Barcode window is still open before quiting the Main program
        Return (cancel the quit) if the Generate Barcode window is still open.
        """
        # Check if generate barcode window is opened
        if self.generate_window_opened:
            # If it is opened show an error
            showerror("Generate Barcode Window is Opened", "Generate Barcode window is still opened!\n"
                                                           "Please close the Generate Barcode window before Quit.")
            return

        # Otherwise check if user want to quit the software
        quit_software = askokcancel("Quit KALMUS", "Are you sure you want to close the KALMUS?\n"
                                                 "All unsaved results will be lost.")
        # Quit if yes
        if quit_software:
            self.quit()

    def quit(self):
        """
        Quit the main window
        """
        self.root.quit()
        self.root.destroy()

    def check_meta_info(self):
        """
        Instantiate the WhichBarcodeCheckMeta window
        """
        WhichBarcodeCheckMeta(self.barcode_1, self.barcode_2, self.barcodes_stack)

    def show_barcode(self):
        """
        Instantiate the WhichBarcodeInspectWindow
        """
        WhichBarcodeInspectWindow(self.barcode_1, self.barcode_2, dpi=100, figsize=(7, 4))

    def load_json_barcode(self):
        """
        Instantiate the LoadJsonWindow
        """
        LoadJsonWindow(self.barcode_gn, self.barcode_1, self.barcode_2, self.ax,
                       self.canvas, self.barcodes_stack)

    def load_stack_barcode(self):
        """
        Instantiate the LoadStackWindow
        """
        LoadStackWindow(self.barcodes_stack, self.barcode_1, self.barcode_2, self.ax,
                        self.canvas)

    def reshape_barcode(self):
        """
        Instantiate the ReshapeBarcodeWindow
        """
        ReshapeBarcodeWindow(self.barcode_1, self.barcode_2, self.ax, self.canvas)

    def stats_info(self):
        """
        Instantiate the StatsInfoWindow
        """
        StatsInfoWindow(self.barcode_1, self.barcode_2)

    def generate_barcode(self):
        """
        Instantiate the GenerateBarcodeWindow
        """
        if not self.generate_window_opened:
            self.generate_window_opened = True
            GenerateBarcodeWindow(self.barcode_gn, self.barcodes_stack)
            self.generate_window_opened = False
        else:
            showinfo("Generate Barcode window is Opened", "Generate Barcode Window is already opened.")

    def save_barcode_on_stack(self):
        """
        Instantiate the SaveBarcodeWindow
        """
        SaveBarcodeWindow(self.barcodes_stack)

    def save_image_from_display(self):
        """
        Instantiate the SaveImageWindow
        """
        SaveImageWindow(self.barcode_1, self.barcode_2)

    def time_pick(self, event):
        """
        Pick the time at a point of the plotted barcode if user double click on that point

        :param event: matplotlib event. Only catch double click event
        """
        # If user is double click on the graph
        if event.dblclick:
            # Try get the x and y position of the clicked point
            try:
                ix, iy = int(event.xdata + 0.5), int(event.ydata + 0.5)
            except Exception:
                return

            # find which axis of the graph does the clicked point belong to
            for i, axe in enumerate(self.ax[:, 0]):
                if axe == event.inaxes:
                    # Check if it is plotted barcode 1 or plotted barcode 2
                    if i == 0:
                        barcode = self.barcode_1
                    else:
                        barcode = self.barcode_2

                    # Make sure the point is within the plotted barcode
                    barcode_shape = barcode.get_barcode().shape
                    if 0 <= iy < barcode_shape[0] and 0 <= ix < barcode_shape[1]:
                        # Instantiate the CheckTimePointWindow
                        CheckTimePointWindow(barcode, mouse_x=ix, mouse_y=iy)
